# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import os
import shutil
import uuid
from collections.abc import Iterable
from typing import Any, Literal

import duckdb
import fsspec
import pandas as pd
import polars as pl
import sqlalchemy as sa
from upath import UPath

import pydiverse.pipedag.backend.table.sql.hooks as sql_hooks
from pydiverse.pipedag import ConfigContext, Schema, Stage, Table
from pydiverse.pipedag.backend.table.sql.ddl import (
    CopySelectTo,
    CreateSchema,
    CreateViewAsSelect,
    DropView,
)
from pydiverse.pipedag.backend.table.sql.dialects.duckdb import DuckDBTableStore
from pydiverse.pipedag.backend.table.sql.sql import DISABLE_DIALECT_REGISTRATION
from pydiverse.pipedag.container import SortOrder, View
from pydiverse.pipedag.context import RunContext
from pydiverse.pipedag.materialize.store import BaseTableStore
from pydiverse.pipedag.materialize.table_hook_base import (
    AutoVersionSupport,
    CanMatResult,
    CanRetResult,
    TableHook,
)
from pydiverse.pipedag.optional_dependency.sqlalchemy import Select, SqlText, TextClause
from pydiverse.pipedag.util.path import is_file_uri


class ParquetTableStore(DuckDBTableStore):
    """Table store that materializes tables as parquet files.

    Additionally, a duckdb file is created which links to parquet files.

    Uses schema swapping for transactions:
    Creates a schema for each stage and a temporary schema for each
    transaction. If all tasks inside a stage succeed, swaps the schemas by
    renaming them.

    .. rubric:: Supported Tables

    The ParquetTableStore can materialize (store task output)
    and dematerialize (retrieve task input) the following Table types:

    .. list-table::
       :widths: 30, 30, 30
       :header-rows: 1
       :stub-columns: 1

       * - Framework
         - Materialization
         - Dematerialization

       * - SQLAlchemy
         - | :py:class:`sa.sql.expression.Selectable <sqlalchemy.sql.expression.Selectable>`
           | :py:class:`sa.sql.expression.TextClause <sqlalchemy.sql.expression.TextClause>`
         - :py:class:`sa.Table <sqlalchemy.schema.Table>`

       * - Pandas
         - :py:class:`pd.DataFrame <pandas.DataFrame>`
         - :py:class:`pd.DataFrame <pandas.DataFrame>`

       * - Polars
         - | :external+pl:doc:`pl.DataFrame <reference/dataframe/index>`
           | :external+pl:doc:`pl.LazyFrame <reference/lazyframe/index>`
         - | :external+pl:doc:`pl.DataFrame <reference/dataframe/index>`
           | :external+pl:doc:`pl.LazyFrame <reference/lazyframe/index>`

       * - tidypolars
         - :py:class:`tp.Tibble <tidypolars.tibble.Tibble>`
         - :py:class:`tp.Tibble <tidypolars.tibble.Tibble>`

       * - Ibis
         - :py:class:`ibis.api.Table <ibis.expr.types.relations.Table>`
         - :py:class:`ibis.api.Table <ibis.expr.types.relations.Table>`

       * - pydiverse.transform
         - ``pdt.Table``
         - | ``pdt.Polars``
           | ``pdt.SqlAlchemy``

       * - pydiverse.pipedag table reference
         - :py:class:`~.ExternalTableReference` (no materialization)
         - Can be read with all dematerialization methods above

       * - pydiverse.pipedag view
         - :py:class:`~.View` (view with support for src union, column renaming, and sorting)
         - Can be read with all dematerialization methods above

    :param url:
        The :external+sa:ref:`SQLAlchemy engine url <database_urls>`
        used to connect to the database.

        This URL may contain placeholders like ``{name}`` or ``{instance_id}``
        (additional ones can be defined in the ``url_attrs_file``) or
        environment variables like ``{$USER}`` which get substituted with their
        respective values.

        Attention: passwords including special characters like ``@`` or ``:`` need
        to be URL encoded.

    :param url_attrs_file:
        Filename of a yaml file which is read shortly before rendering the final
        engine URL and which is used to replace custom placeholders in ``url``.

        Just like ``url``, this value may also contain placeholders and environment
        variables which get substituted.

    :param create_database_if_not_exists:
        If the engine url references a database name that doesn't yet exists,
        then setting this value to ``True`` tells pipedag to create the database
        before trying to open a connection to it.

    :param schema_prefix:
        Prefix placed in front of all schema names created by pipedag.
    :param schema_suffix:
        Suffix placed behind all schema names created by pipedag.
    :param avoid_drop_create_schema:
        If ``True``, no ``CREATE SCHEMA`` or ``DROP SCHEMA`` statements get issued.
        This is mostly relevant for databases that support automatic schema
        creation like IBM DB2.

    :param print_materialize:
        If ``True``, log every table materialization.
    :param print_sql:
        If ``True``, log every executed SQL statement.
    :param no_db_locking:
        Speed up database by telling it we will not rely on it's locking mechanisms.
        Currently not implemented.

    :param strict_materialization_details:
        If ``True``: raise an exception if
            - the argument ``materialization_details`` is given even though the
              table store does not support it.
            - a table references a ``materialization_details`` tag that is not defined
              in the config.
        If ``False``: Log an error instead of raising an exception

    :param materialization_details:
        A dictionary with each entry describing a tag for materialization details of
        the table store. See subclasses of :py:class:`BaseMaterializationDetails`
        for details.
    :param default_materialization_details:
        The materialization_details that will be used if materialization_details
        is not specified on table level. If not set, the ``__any__`` tag (if specified)
        will be used.
    :param max_concurrent_copy_operations:
        In case of a partially cache-valid stage, we need to copy tables from the
        cache schema to the new transaction schema. This parameter specifies the
        maximum number of workers we use for concurrently copying tables.
    :param sqlalchemy_pool_size:
        The number of connections to keep open inside the connection pool.
        It is recommended to choose a larger number than
        ``max_concurrent_copy_operations`` to avoid running into pool_timeout.
    :param sqlalchemy_pool_timeout:
        The number of seconds to wait before giving up on getting a connection from
        the pool. This may be relevant in case the connection pool is saturated
        with concurrent operations each working with one or more database connections.
    :param force_transaction_suffix:
        This option is for use without proper pipedag flow. It disables lookup in stage
        metadata table for determining correct transaction slot suffix.
        (default: None)
    :param parquet_base_bath:
        This is an fsspec compatible path to either a directory or an S3 bucket key
        prefix. Examples are ``/tmp/pipedag/parquet/`` or ``s3://pipedag-test-bucket/table_store/``.
        ``instance_id`` will automatically be appended to parquet_base_bath.
    :param s3_endpoint_url:
        When using a non-standard S3 endpoint (like minio), this can be used to specify the URL.
        Unfortunately, the AWS_ENDPOINT_URL environment variable is not automatically picked up by duckdb.
        (default: None)
    :param s3_use_ssl:
        Whether to use SSL when connecting to S3 endpoint. (default: None)
    :param s3_url_style:
        Specify URL style when connecting to S3 endpoint. Minio for example requires s3_url_style='path'.
        (default: None)
    :param allow_overwrite:
        Allow overwriting tables with store_table.
        (default: False)
    """

    _dialect_name = DISABLE_DIALECT_REGISTRATION

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        config = config.copy()
        config["parquet_base_path"] = UPath(config["parquet_base_path"])
        return super()._init_conf_(config)

    def __init__(
        self,
        engine_url: str,
        parquet_base_path: UPath,
        *,
        create_database_if_not_exists: bool = False,
        schema_prefix: str = "",
        schema_suffix: str = "",
        avoid_drop_create_schema: bool = False,
        print_materialize: bool = False,
        print_sql: bool = False,
        no_db_locking: bool = True,
        strict_materialization_details: bool = True,
        materialization_details: dict[str, dict[str | list[str]]] | None = None,
        default_materialization_details: str | None = None,
        max_concurrent_copy_operations: int = 5,
        sqlalchemy_pool_size: int = 12,
        sqlalchemy_pool_timeout: int = 300,
        force_transaction_suffix: str | None = None,
        s3_endpoint_url: str | None = None,
        s3_url_style: str | None = None,
        s3_region: str | None = None,
        allow_overwrite: bool = False,
    ):
        # must be initialized before super().__init__() call because self._create_engine() is called there
        self.parquet_base_path = parquet_base_path
        self.s3_endpoint_url = s3_endpoint_url
        self.s3_url_style = s3_url_style
        self.s3_region = s3_region
        self.allow_overwrite = allow_overwrite
        super().__init__(
            engine_url,
            create_database_if_not_exists=create_database_if_not_exists,
            schema_prefix=schema_prefix,
            schema_suffix=schema_suffix,
            avoid_drop_create_schema=avoid_drop_create_schema,
            print_materialize=print_materialize,
            print_sql=print_sql,
            no_db_locking=no_db_locking,
            strict_materialization_details=strict_materialization_details,
            materialization_details=materialization_details,
            default_materialization_details=default_materialization_details,
            max_concurrent_copy_operations=max_concurrent_copy_operations,
            sqlalchemy_pool_size=sqlalchemy_pool_size,
            sqlalchemy_pool_timeout=sqlalchemy_pool_timeout,
            force_transaction_suffix=force_transaction_suffix,
        )

        # Add parquet specific metadata table for user id.
        # This can be UUID which is constant per .duckdb file and used to sync with metadata_store.
        self.sql_metadata_local = sa.MetaData(schema=self.metadata_schema.get())
        self.user_id_table = sa.Table(
            "user_id",
            self.sql_metadata_local,
            sa.Column("user_id", sa.String(64)),
        )
        self.sync_views_table = None

        # ## state

        # whether parquet files of stage are stored in __odd or __even
        # can typically be found in metadata, but firstly we cache this
        # and secondly we modify it for the current stage before commit
        self.parquet_schema_paths: dict[str, UPath] = {}
        # tables may get alias to other transaction slot in case of cache
        # valid tasks
        self.parquet_table_paths: dict[tuple[str, str], UPath] = {}
        # remember tables that need to be copied if stage is not unchanged
        self.parquet_deferred_copy: list[Table] = []

        self.logger.info(
            "Initialized Parquet Table Store",
            parquet_base_path=self.parquet_base_path,
            engine_url=self.engine_url.render_as_string(hide_password=True),
            schema_prefix=self.schema_prefix,
            schema_suffix=self.schema_suffix,
        )

    def set_metadata_store(self, store: BaseTableStore):
        super().set_metadata_store(store)

        # make sure sync_views_table exists in metadata_store
        self.sync_views_table = sa.Table(
            "sync_views",
            store.sql_metadata,
            sa.Column("schema", sa.String(128), primary_key=True),
            sa.Column("view_name", sa.String(128), primary_key=True),
            sa.Column("user_id", sa.String(64), primary_key=True),
            sa.Column("target", sa.String(1024)),
            sa.Column("target_type", sa.String(64)),
            sa.Column("obsolete", sa.SmallInteger()),  # some databases have problems with bool
        )

    def setup(self):
        super().setup()

        if self.metadata_store:
            # super().setup() already called self.metadata_store.setup() and
            # self.metadata_store.sql_metadata.create_all() incl. sync_views_table

            self.sql_metadata_local.create_all(self.engine, checkfirst=True)
            with self.engine_connect() as conn:
                self.user_id = conn.execute(sa.select(self.user_id_table.c.user_id)).scalar_one_or_none()
                if self.user_id is None:
                    # create a new UUID if it does not exist, yet
                    self.user_id = uuid.uuid4().hex
                    conn.execute(sa.insert(self.user_id_table).values(user_id=self.user_id))

    def get_storage_options(self, kind: Literal["polars", "fsspec"], protocol) -> dict[str, Any] | None:
        if protocol != "s3" or (self.s3_endpoint_url is None and self.s3_region is None):
            return None
        if kind == "polars":
            return {
                k: v
                for k, v in dict(
                    aws_endpoint_url=self.s3_endpoint_url,
                    aws_region=self.s3_region,
                    addressing_style=self.s3_url_style,
                ).items()
                if v is not None
            }
        elif kind == "fsspec":
            return {
                "client_kwargs": {
                    "endpoint_url": self.s3_endpoint_url,
                    "region_name": self.s3_region,
                },
                "config_kwargs": {"s3": {"addressing_style": self.s3_url_style}},
            }

    def _create_engine(self):
        engine = super()._create_engine()

        def execute(con, statement):
            """
            Execute a SQL statement and return the result.
            """
            if sa.__version__ >= "2.0.0":
                return con.execute(sa.text(statement))
            else:
                return con.execute(statement)

        with engine.connect() as con:
            if self.parquet_base_path.protocol == "s3":
                # Unfortunately it is hard to configure S3 URL endpoint consistently for fsspec, duckdb, and polars.
                # Thus, we take it as parameter to ParquetTableStore and pass it to fsspec. Fsspec in turn
                # is passed on to duckdb and polars. This is only needed for non-standard S3 endpoints like minio.
                from fsspec.config import conf  # fsspec’s global config

                if "s3" not in conf and self.s3_endpoint_url is not None:
                    conf.setdefault("s3", {})
                    conf["s3"].update(
                        {
                            "client_kwargs": {
                                "endpoint_url": self.s3_endpoint_url,
                            },
                            # ─── botocore.config.Config parameters go here ────
                            "config_kwargs": {
                                "connect_timeout": 3,
                                "read_timeout": 5,
                                "retries": {"max_attempts": 2, "mode": "standard"},
                            },
                        }
                    )
                    if self.s3_region:
                        conf["s3"]["client_kwargs"]["region_name"] = self.s3_region
                    if self.s3_url_style:
                        # MinIO needs path-style
                        conf["s3"]["config_kwargs"]["s3"] = dict(addressing_style=self.s3_url_style)

            if sa.__version__ >= "2.0.0":
                con.commit()

        # This is needed for gcsfs. For s3fs, it would also be possible to
        # go with httpfs (https://duckdb.org/docs/stable/core_extensions/httpfs/s3api.html).
        fs = fsspec.filesystem(self.parquet_base_path.protocol)
        engine.raw_connection().register_filesystem(fs)

        return engine

    def metadata_sync_views(self, schema_name: str):
        """
        Sync views between different users via metadata_store.

        This function interacts with metadata_track_* functions to get input in the metadata table sync_views.

        The per-user .duckdb file stores a sync_views_user_id table:
        - user_id  # UUID identifying the user taking part in the synchronization

        The sync_views table holds the following information:
        - schema [PK]
        - view_name [PK]
        - user_id [PK]
        - target  # parquet_file or target_schema
        - target_type  # "parquet": target is a parquet_file, "schema": target is a redirect schema name
        - obsolete  # if true (obsolete != 0), the parquet file is not in the correct place any more and
                    # the view must be removed

        Logic:
        - any schema reset will remove parquet files, delete all entries for schema
        - any view creation will insert or update the respective entry for the respective user_id; it will also update
        all entries by other user_ids to obsolete=True;
        - any view deletion will remove the respective entry for this user_id, and update all entries for other user_id
        values to obsolete=True
        - metadata_sync_views() will delete all views in the current schema if for the user_id, there is no entry for
        this schema. If there is already an entry for this user_id, it will delete all views in the current schema
          which are marked obsolete=True for this user_id, and it will create all views which are obsolete=False and
          from another user.
        """
        if self.metadata_store:
            with self.metadata_store.engine_connect() as meta_conn:
                tbl = self.sync_views_table
                match_user_id = (tbl.c.schema == schema_name) & (tbl.c.user_id == self.user_id)
                obsolete_views = meta_conn.execute(
                    sa.select(tbl.c.view_name).where(match_user_id & (tbl.c.obsolete != 0))
                ).fetchall()
                if len(obsolete_views) > 0:
                    self.logger.info(
                        "Removing obsolete views due to sync with metadata_store",
                        schema=schema_name,
                        views=[v[0] for v in obsolete_views],
                    )
                for view in obsolete_views:
                    try:
                        meta_conn.execute(DropView(view[0], schema_name, if_exists=False))
                    except sa.exc.OperationalError:
                        self.log.error(
                            f"Drop view failed while reconciling views with metadata_store: {schema_name}.{view[0]}"
                        )
                deleted = meta_conn.execute(tbl.delete().where(match_user_id & (tbl.c.obsolete != 0))).rowcount
                if deleted > 0:
                    self.logger.info(
                        "Deleted obsolete view entries in metadata_store",
                        schema=schema_name,
                        user_id=self.user_id,
                        affected=deleted,
                    )
                if len(obsolete_views) == 0:
                    user_id_cnt = meta_conn.execute(
                        sa.select(sa.func.count(sa.text("*"))).where(match_user_id)
                    ).fetchall()[0][0]
                    if user_id_cnt == 0:
                        # this user_id has no entry for this schema -> remove all views in this schema
                        with self.engine_connect() as conn:
                            existing_views = conn.execute(
                                sa.text(f"FROM duckdb_views() SELECT view_name WHERE schema_name='{schema_name}'")
                            ).fetchall()
                            if len(existing_views) > 0:
                                self.logger.info(
                                    "Removing all views due to sync with metadata_store",
                                    schema=schema_name,
                                    views=[v[0] for v in existing_views],
                                )
                            for view in existing_views:
                                try:
                                    conn.execute(DropView(view[0], schema_name, if_exists=False))
                                except sa.exc.OperationalError:
                                    self.log.error(
                                        "Drop view failed while reconciling views with metadata_store: "
                                        f"{schema_name}.{view[0]}"
                                    )

                query = sa.select(tbl.c.view_name, tbl.c.target, tbl.c.target_type).where(
                    (tbl.c.schema == schema_name) & (tbl.c.user_id != self.user_id) & (tbl.c.obsolete == 0)
                )
                create_views = meta_conn.execute(query).fetchall()
                if len(create_views) > 0:
                    self.logger.info(
                        "Creating views due to sync with metadata_store",
                        schema=schema_name,
                        views=[v[0] for v in create_views],
                    )
                with self.engine_connect() as conn:
                    del meta_conn  # prevent typo errors
                    for view, target, target_type in create_views:
                        try:
                            conn.execute(DropView(view, schema_name, if_exists=True))
                            schema = Schema(schema_name, prefix="", suffix="")
                            if target_type == "parquet":
                                conn.execute(CreateViewAsSelect(view, schema, self._read_parquet_query(target)))
                            elif target_type == "schema":
                                conn.execute(
                                    CreateViewAsSelect(view, schema, sa.text(f"SELECT * FROM {target}.{view}"))
                                )
                            else:
                                meta_engine = self.metadata_schema.engine
                                self.log.error(
                                    f"Unknown target_type in sync_views table: {target_type}\n"
                                    f"{query.compile(meta_engine, compile_kwargs={'literal_binds': True})}"
                                )
                        except sa.exc.OperationalError:
                            self.log.error(
                                "Create view failed while reconciling views with metadata_store: "
                                f"{schema_name}.{view} -> {target} ({target_type})"
                            )

    def metadata_track_view_flush(self, schema_name: str):
        """
        Flush all entries for schema in metadata_store.

        see metadata_sync_views() for details.
        """
        if self.metadata_store:
            with self.metadata_store.engine_connect() as meta_conn:
                tbl = self.sync_views_table
                deleted = meta_conn.execute(tbl.delete().where(tbl.c.schema == schema_name)).rowcount
                if deleted > 0:
                    self.logger.info(
                        "Flushed all view entries in metadata_store for schema", schema=schema_name, deleted=deleted
                    )

    def metadata_track_view(
        self, table_name: str, schema_name: str, target: str, target_type: Literal["parquet", "schema"]
    ):
        """
        Track creation of a view in metadata_store.

        see metadata_sync_views() for details.
        """
        if self.metadata_store:
            with self.metadata_store.engine_connect() as meta_conn:
                tbl = self.sync_views_table
                view_match = (tbl.c.schema == schema_name) & (tbl.c.view_name == table_name)
                # mark all other user views as obsolete
                updated = meta_conn.execute(tbl.update().where(view_match).values(obsolete=1)).rowcount
                if updated > 0:
                    self.logger.info(
                        "Marked views by other users as obsolete in metadata_store",
                        schema=schema_name,
                        view=table_name,
                        affected=updated,
                    )
                # drop obsolete own match:
                deleted = meta_conn.execute(tbl.delete().where(view_match & (tbl.c.user_id == self.user_id))).rowcount
                if deleted > 0:
                    self.logger.info(
                        "Deleted obsolete own view entries in metadata_store",
                        schema=schema_name,
                        view=table_name,
                        user_id=self.user_id,
                        deleted=deleted,
                    )
                # insert view
                meta_conn.execute(
                    sa.insert(tbl).values(
                        schema=schema_name,
                        view_name=table_name,
                        user_id=self.user_id,
                        target=target,
                        target_type=target_type,
                        obsolete=0,
                    )
                )
                self.logger.info(
                    "Inserted view entry in metadata_store",
                    schema=schema_name,
                    view=table_name,
                    target=target,
                    target_type=target_type,
                )

    def metadata_track_view_drop(self, table_name: str, schema_name: str):
        """
        Track deletion of a view in metadata_store.

        see metadata_sync_views() for details.
        """
        if self.metadata_store:
            with self.metadata_store.engine_connect() as meta_conn:
                tbl = self.sync_views_table
                view_match = (tbl.c.schema == schema_name) & (tbl.c.view_name == table_name)
                deleted = meta_conn.execute(tbl.delete().where(view_match & (tbl.c.user_id == self.user_id))).rowcount
                if deleted > 0:
                    self.logger.info(
                        "Deleted view entry in metadata_store", schema=schema_name, view=table_name, deleted=deleted
                    )
                # mark all other user views as obsolete
                updated = meta_conn.execute(tbl.update().where(view_match).values(obsolete=1)).rowcount
                if updated > 0:
                    self.logger.info(
                        "Marked views by other users as obsolete in metadata_store",
                        schema=schema_name,
                        view=table_name,
                        affected=updated,
                    )

    def on_clear_schema(self, schema_name: str):
        # this callback is called by parent class in _commit_stage_read_views()
        self.metadata_track_view_flush(schema_name)

    def on_read_view_alias(self, table_name: str, from_schema: str, to_schema: str):
        # this callback is called by parent class in _commit_stage_read_views()
        self.metadata_track_view(table_name, to_schema, target=from_schema, target_type="schema")

    def init_stage(self, stage: Stage):
        # fetch existing tables from database before schema is deleted there
        old_transaction_name = self._get_read_view_transaction_name(stage.name)
        new_transaction_name = self._get_read_view_original_transaction_name(stage, old_transaction_name)

        # first synchronize views to match the actual state in metadata_store
        if self.metadata_store:
            # Typically the parent schema stage.name references old_transaction schema.
            # But we also need to sync the new transaction schema to avoid double deletion of parquet files.
            schemas = [old_transaction_name] if old_transaction_name != "" else []
            schemas = schemas + [self.get_schema(stage.name).get(), new_transaction_name]
            for schema in schemas:
                self.execute(CreateSchema(Schema(schema, prefix="", suffix=""), if_not_exists=True))
                self.metadata_sync_views(schema)

        with self.engine_connect() as conn:
            existing_tables = self.execute(
                f"FROM duckdb_views() SELECT view_name WHERE "
                f"schema_name='{new_transaction_name}' "
                f"and sql like '%%FROM read_parquet(%%'",
                conn=conn,
            ).fetchall()
        # This creates transaction schema and modifies stage.current_name.
        # It also removes all views that were previously in transaction schema.
        super().init_stage(stage)
        self.metadata_track_view_flush(new_transaction_name)
        # update cache: this is important since before this stage commits,
        # self._get_read_view_transaction_name() will yield the wrong name
        new_schema = self.get_schema(stage.current_name)
        self.parquet_schema_paths[new_schema.name] = self.get_parquet_path(new_schema)
        # The stage.name should still point to opposite transaction schema
        old_schema = self.get_schema(self._get_read_view_original_transaction_name(stage))
        self.parquet_schema_paths[self.get_schema(stage.name).name] = self.get_parquet_path(old_schema)
        # TODO: for nested stages this might be per stage
        self.parquet_deferred_copy.clear()
        path = self.get_stage_path(stage)
        if is_file_uri(path):
            os.makedirs(path, exist_ok=True)
        for row in existing_tables:
            table = row[0]
            file_path = path / (table + ".parquet")
            try:
                # import traceback
                self.logger.info(
                    "Cleaning up parquet file in transaction schema",
                    file_path=file_path,
                    # stacktrace="\n"+"\n".join(traceback.format_stack()),
                )
                file_path.unlink()
            except FileNotFoundError:
                self.logger.error(
                    "Could not remove parquet file while deleting corresponding view in transaction schema",
                    file=file_path,
                    view_name=table,
                    stage=stage,
                )

    def _copy_table(self, table: Table, from_schema: Schema, from_name: str):
        if table.name != from_name:
            # we ignore the _copy_table which wants to write __copy tables
            return
        _ = from_schema  # not used because this only points to READ VIEW
        dest_schema = self.get_schema(table.stage.current_name)
        original_transaction_name = self._get_read_view_original_transaction_name(table.stage)
        original_schema = self.get_schema(original_transaction_name)
        src_file_path = self.get_parquet_path(original_schema) / (from_name + ".parquet")
        dest_file_path = self.get_parquet_path(dest_schema) / src_file_path.name
        self.logger.info(
            "Copying table between transactions",
            src_file_path=src_file_path,
            dest_file_path=dest_file_path,
        )
        if is_file_uri(src_file_path):
            shutil.copy(src_file_path, dest_file_path)
        else:
            src_file_path.fs.copy(src_file_path.as_uri(), dest_file_path.as_uri(), on_error="raise")
        # create view in duckdb database file
        self.execute(CreateViewAsSelect(table.name, dest_schema, self._read_parquet_query(dest_file_path)))
        self.metadata_track_view(table.name, dest_schema.get(), dest_file_path.as_uri(), "parquet")

    @staticmethod
    def _read_parquet_query(file_path):
        # attention: we filter for %%FROM read_parquet(%% in init_stage
        return sa.text(f"FROM read_parquet('{file_path}')")

    def _swap_alias_with_table_copy(self, table: Table, table_copy: Table):
        # There is no __copy table which can be renamed for this table store.
        # Alias will be replaced in commit_stage()
        pass

    def _deferred_copy_table(
        self,
        table: Table,
        from_schema: Schema,
        from_name: str,
    ):
        # this will just create an alias in the duckdb database
        super()._deferred_copy_table(table, from_schema, from_name)
        # keep symlink to original parquet file until stage commits
        original_transaction_name = self._get_read_view_original_transaction_name(table.stage)
        original_schema = self.get_schema(original_transaction_name)
        schema = self.get_schema(table.stage.current_name)
        self.parquet_table_paths[(schema.name, table.name)] = self.get_parquet_path(original_schema) / (
            table.name + ".parquet"
        )
        self.parquet_deferred_copy.append(table)

    def commit_stage(self, stage: Stage):
        schema = self.get_schema(stage.current_name)
        stage_transaction_path = self.parquet_schema_paths[schema.name]
        # deferred copy of parquet files if stage is not 100% cache valid
        if RunContext.get().has_stage_changed(stage):
            for table in self.parquet_deferred_copy:
                # remove views which were placed as aliases
                self.execute(DropView(table.name, schema))
                src_file_path = self.parquet_table_paths[(schema.name, table.name)]
                dest_file_path = stage_transaction_path / src_file_path.name
                self.logger.info(
                    "Copying table between transactions",
                    src_file_path=src_file_path,
                    dest_file_path=dest_file_path,
                )
                if is_file_uri(src_file_path):
                    shutil.copy(src_file_path, dest_file_path)
                else:
                    src_file_path.fs.copy(src_file_path.as_uri(), dest_file_path.as_uri(), on_error="raise")
                # replace view in duckdb database file (which was previously just
                # alias to main schema)
                self.execute(CreateViewAsSelect(table.name, schema, self._read_parquet_query(dest_file_path)))
                self.metadata_track_view(table.name, schema.get(), dest_file_path.as_uri(), "parquet")
            # switch committed transaction path to the new transaction path
            self.parquet_schema_paths[self.get_schema(stage.name).name] = stage_transaction_path
        super().commit_stage(stage)
        # clear table individual parquet paths since this is not needed after commit
        self.parquet_table_paths = {}

    def _committed_unchanged(self, stage):
        super()._committed_unchanged(stage)
        # transaction schema is discarded, so we don't need to copy tables there
        self.parquet_deferred_copy.clear()

    def get_parquet_path(self, schema: Schema):
        return self.parquet_base_path / self.instance_id / schema.get()

    def get_parquet_schema_path(self, schema: Schema) -> UPath:
        # Parquet files are stored in transaction schema while stage.current_name
        # changes to final schema. We resolve a level of indirection in memory.
        if schema.name in self.parquet_schema_paths:
            return self.parquet_schema_paths[schema.name]
        # now we assume schema.name == stage.name (any stage with
        # current_name != name should be in self.parquet_schema_paths)
        transaction_name = self._get_read_view_transaction_name(schema.name)
        if transaction_name == "":
            transaction_name = schema.name
        path = self.get_parquet_path(self.get_schema(transaction_name))
        self.parquet_schema_paths[schema.name] = path
        return path

    def get_stage_path(self, stage: Stage) -> UPath:
        try:
            cfg = ConfigContext.get()
        except LookupError:
            # schema-prefix/suffix not available if no ConfigContext active
            return self.get_parquet_schema_path(Schema(stage.current_name))
        store = cfg.store.table_store
        return self.get_parquet_schema_path(store.get_schema(stage.current_name))

    def get_table_path(
        self,
        table: Table,
        file_extension: str = ".parquet",
    ) -> UPath:
        schema_name = table.stage.current_name
        try:
            cfg = ConfigContext.get()
            store = cfg.store.table_store
            schema = store.get_schema(schema_name)
        except LookupError:
            # schema-prefix/suffix not available if no ConfigContext active
            schema = Schema(schema_name)
        return self.get_table_schema_path(table.name, schema, file_extension)

    def get_table_schema_path(self, table_name: str, schema: Schema, file_extension: str = ".parquet") -> UPath:
        # Parquet files might be stored in other transaction schema in case of cache
        # validity. We resolve a level of indirection in memory.
        if (schema.name, table_name) in self.parquet_table_paths:
            return self.parquet_table_paths[(schema.name, table_name)]
        return self.get_parquet_schema_path(schema) / (table_name + file_extension)

    def delete_table_from_transaction(self, table: Table, *, schema: Schema | None = None):
        if schema is None:
            schema = self.get_schema(table.stage.transaction_name)
        file_path = self.get_table_schema_path(table.name, schema)
        file_path.unlink(missing_ok=True)
        view_schema = self.get_schema(table.stage.transaction_name)
        self.execute(
            DropView(
                table.name,
                view_schema,
                if_exists=True,
            )
        )
        self.metadata_track_view_drop(table.name, view_schema.get())

    def add_primary_key(
        self,
        table_name: str,
        schema: Schema,
        key_columns: list[str],
        *,
        name: str | None = None,
    ):
        # ParquetTableStore does not support indexes. They are ignored.
        pass

    def add_index(
        self,
        table_name: str,
        schema: Schema,
        index_columns: list[str],
        name: str | None = None,
    ):
        # ParquetTableStore does not support indexes. They are ignored.
        pass

    def copy_indexes(self, src_table: str, src_schema: Schema, dest_table: str, dest_schema: Schema):
        # ParquetTableStore does not support indexes. They are ignored.
        pass

    def rename_table(self, table: Table, to_name: str, schema: Schema):
        _ = schema  # ignore given schema
        schema = self.get_schema(table.stage.transaction_name)
        from_path = self.get_table_schema_path(table.name, schema)
        to_path = self.get_table_schema_path(to_name, schema)
        if is_file_uri(from_path):
            os.rename(from_path, to_path)
        else:
            from_path.fs.copy(from_path.as_uri(), to_path.as_uri(), on_error="raise")
            from_path.unlink()

        # drop view
        self.execute(
            DropView(
                table.name,
                schema,
            )
        )
        self.metadata_track_view_drop(table.name, schema.get())
        # create view again
        self.execute(CreateViewAsSelect(to_name, schema, self._read_parquet_query(to_path)))
        self.metadata_track_view(to_name, schema.get(), to_path.as_uri(), "parquet")

    def write_subquery(
        self,
        query,
        to_name,
        neighbor_table: Table,
        *,
        unlogged: bool = False,
        suffix: str | None = None,
    ):
        """Write a query to a table in same schema as neighbor_table.

        This is mainly for overriding by derived table stores like ParquetTableStore.
        """
        _ = unlogged, suffix  # materialization details not used by ParquetTableStore
        schema = self.get_schema(neighbor_table.stage.transaction_name)
        file_path = self.get_table_schema_path(to_name, schema)
        self.execute(
            [
                CopySelectTo(
                    file_path,
                    "PARQUET",
                    query,
                ),
                CreateViewAsSelect(to_name, schema, self._read_parquet_query(file_path)),
            ]
        )
        self.metadata_track_view(to_name, schema.get(), file_path.as_uri(), "parquet")
        return schema

    def drop_subquery_table(
        self,
        drop_name: str,
        schema: Schema | str,
        neighbor_table: Table,
        if_exists: bool = False,
        cascade: bool = False,
    ):
        """Drop a table in same schema as neighbor_table.

        This is mainly for overriding by derived table stores like ParquetTableStore.
        """
        _ = schema  # ignore given schema; use neighbor_table schema instead
        _ = cascade
        schema = self.get_schema(neighbor_table.stage.transaction_name)
        drop_path = self.get_table_schema_path(drop_name, schema)
        try:
            drop_path.unlink()
        except FileNotFoundError as e:
            if not if_exists:
                raise e
        # drop view
        self.execute(
            DropView(
                drop_name,
                schema,
            )
        )
        self.metadata_track_view_drop(drop_name, schema.get())


@ParquetTableStore.register_table(pd)
class PandasTableHook(TableHook[ParquetTableStore]):
    auto_version_support = AutoVersionSupport.TRACE

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, pd.DataFrame))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(issubclass(type_, pd.DataFrame))

    @classmethod
    def materialize(
        cls,
        store: ParquetTableStore,
        table: Table[pd.DataFrame],
        stage_name: str,
        without_config_context: bool = False,
    ):
        file_path = store.get_table_path(table)
        schema = store.get_schema(table.stage.current_name)

        if store.print_materialize:
            store.logger.info(
                f"Writing pandas table '{schema.get()}.{table.name}'",
                file_path=file_path,
                table_obj=table.obj,
            )

        df = table.obj
        df.to_parquet(
            str(file_path), index=False, storage_options=store.get_storage_options("fsspec", file_path.protocol)
        )
        store.execute(CreateViewAsSelect(table.name, schema, store._read_parquet_query(file_path)))
        store.metadata_track_view(table.name, schema.get(), file_path.as_uri(), "parquet")

    @classmethod
    def retrieve(
        cls,
        store: ParquetTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type,
        limit: int | None = None,
    ):
        # view resolution in input table is implemented here using pyarrow

        path = store.get_table_path(table)
        import pyarrow.dataset as ds

        pyarrow_path, pyarrow_fs = cls.get_pyarrow_path(path, store)
        first_pyarrow_path = pyarrow_path

        if table.view:
            view = table.view
            assert view.assert_normalized
            if isinstance(view.src, Iterable):
                path = [store.get_table_path(tbl) for tbl in view.src]
                pyarrow_path = [cls.get_pyarrow_path(p, store)[0] for p in path]
                first_pyarrow_path = pyarrow_path[0]
            else:
                path = store.get_table_path(view.src)
                pyarrow_path, _ = cls.get_pyarrow_path(path, store)
                first_pyarrow_path = pyarrow_path

            ads = ds.dataset(pyarrow_path, filesystem=pyarrow_fs)
            if view.sort_by is not None:
                # Unfortunately this loads the full table with all columns in memory.
                # Calling PolarsTableHook might be more efficient if a dataset is large and many
                # columns are dropped in view.columns.
                sort_cols = view.sort_by if isinstance(view.sort_by, Iterable) else [view.sort_by]
                ads = ads.sort_by(
                    [
                        (sort_col.col, "descending" if sort_col.order == SortOrder.DESC else "ascending")
                        for sort_col in sort_cols
                    ]
                )
            import pyarrow.compute as pc

            ads = ads.scanner(columns={k: pc.field(v) for k, v in view.columns.items()} if view.columns else None)
            if view.limit is not None:
                ads = ads.head(view.limit)
            else:
                ads = ads.to_table()  # whole table (needed for to_pandas below)
            df = ads.to_pandas()
            schema_names, schema_types = cls.get_pyarrow_schema(first_pyarrow_path, pyarrow_fs)
            if view.columns:
                # apply column transformation to pyarrow schema
                schema_in = {k: v for k, v in zip(schema_names, schema_types, strict=True)}
                schema_out = {k: schema_in[v] for k, v in view.columns.items()}
                schema_names = schema_out.keys()
                schema_types = schema_out.values()
        else:
            if limit is not None:
                df = ds.dataset(pyarrow_path, filesystem=pyarrow_fs).scanner().head(limit).to_pandas()
            else:
                df = pd.read_parquet(str(path), storage_options=store.get_storage_options("fsspec", path.protocol))
            schema_names, schema_types = cls.get_pyarrow_schema(first_pyarrow_path, pyarrow_fs)
        df = df.astype(
            {
                col: "datetime64[s]"
                for col, type_, dtype in zip(schema_names, schema_types, df.dtypes, strict=True)
                if type_ == "date32" and dtype == object  # noqa: E721
            }
        )
        if isinstance(as_type, tuple) and len(as_type) == 2 and len(schema_names) > 0:
            if as_type[1] == "arrow" and not all(
                hasattr(dtype, "storage") and dtype.storage == "pyarrow" for dtype in df.dtypes
            ):
                store.logger.warning(
                    f"Ignoring storage specialization '{as_type[1]}' for {table.name} "
                    f"as ParquetTableStore reads pandas the same encoding as it writes"
                    f" it. This can be changed if need arises.",
                    as_type=as_type,
                    dtypes=df.dtypes,
                )
            if as_type[1] == "numpy" and any(
                hasattr(dtype, "storage") and dtype.storage == "pyarrow" for dtype in df.dtypes
            ):
                store.logger.warning(
                    f"Ignoring storage specialization '{as_type[1]}' for {table.name} "
                    f"as ParquetTableStore reads pandas the same encoding as it writes"
                    f" it. This can be changed if need arises.",
                    as_type=as_type,
                    dtypes=df.dtypes,
                )
        if table.name is not None:
            df.attrs["name"] = table.name
        return df

    @classmethod
    def get_pyarrow_schema(cls, path: str, pyarrow_fs: fsspec.AbstractFileSystem) -> tuple[Any, Any]:
        import pyarrow.parquet

        # attention: with categorical columns, it might be necessary to fuse dictionaries of all parquet files
        schema = pyarrow.parquet.read_schema(path, filesystem=pyarrow_fs)
        schema_names, schema_types = schema.names, schema.types
        return schema_names, schema_types

    @staticmethod
    def get_pyarrow_path(path: UPath, store: ParquetTableStore) -> tuple[str, fsspec.AbstractFileSystem]:
        if path.protocol == "s3" and (store.s3_endpoint_url or store.s3_region):
            # Setting endpoint URL is different for fsspec, pyarrow, duckdb, polars, and pandas.
            # ParquetTableStore allows setting endpoint_url in pipedag.yaml and copies it to fsspec and duckdb.
            # Here, we copy it from fsspec to pyarrow.
            pyarrow_path = path.path
            # unfortunately, force_virtual_addressing=store.s3_url_style=="path" does not work for
            # pyarrow.parquet.read_schema(pyarrow_path, filesystem=pyarrow_fs)
            import pyarrow.fs

            pyarrow_fs = pyarrow.fs.S3FileSystem(endpoint_override=store.s3_endpoint_url, region=store.s3_region)
        else:
            pyarrow_path = path
            pyarrow_fs = None
        return pyarrow_path, pyarrow_fs

    @classmethod
    def auto_table(cls, obj: pd.DataFrame):
        return sql_hooks.PandasTableHook.auto_table(obj)

    @classmethod
    def get_computation_tracer(cls):
        return sql_hooks.PandasTableHook.ComputationTracer()


@ParquetTableStore.register_table(pl, duckdb)
class PolarsTableHook(sql_hooks.PolarsTableHook):
    def _execute_materialize_polars(table, store, stage_name):
        _ = stage_name
        file_path = store.get_table_path(table)
        schema = store.get_schema(table.stage.current_name)
        df = table.obj

        if store.print_materialize:
            store.logger.info(
                f"Writing polars table '{schema.get()}.{table.name}' to parquet",
                file_path=file_path,
            )
        if options := store.get_storage_options("polars", file_path.protocol):
            if pl.__version__ < "1.3.0":
                raise RuntimeError(
                    "Storing polars tables with custom storage options is not supported for polars < 1.3.0. "
                    f"Current version is {pl.__version__}: {options}"
                )
            # at some point polars supported UPath, but 1.33.1 does not
            df.write_parquet(str(file_path), storage_options=options)
        else:
            df.write_parquet(str(file_path))
        if store.allow_overwrite:
            store.execute(DropView(table.name, schema, if_exists=True))
        store.execute(CreateViewAsSelect(table.name, schema, store._read_parquet_query(file_path)))
        store.metadata_track_view(table.name, schema.get(), file_path.as_uri(), "parquet")

    @classmethod
    def _execute_query(
        cls,
        store: ParquetTableStore,
        table: Table,
        stage_name: str,
        as_type: type,
        limit: int | None = None,
    ) -> pl.DataFrame:
        _ = as_type
        if table.view:
            view = table.view
            if isinstance(view.src, Iterable):
                file_paths = [store.get_table_path(tbl) for tbl in view.src]
                protocol = file_paths[0].protocol
                file_paths = [str(path) for path in file_paths]
            else:
                file_paths = store.get_table_path(view.src)
                protocol = file_paths.protocol
                file_paths = str(file_paths)
            lf = pl.scan_parquet(
                file_paths, n_rows=limit, storage_options=store.get_storage_options("polars", protocol)
            )
            if view.sort_by is not None:
                sort_cols = [c.col for c in view.sort_by]
                sort_desc = [c.order == SortOrder.DESC for c in view.sort_by]
                sort_nulls_last = [c.nulls_first == False for c in view.sort_by]  # noqa: E712
                lf = lf.sort(sort_cols, descending=sort_desc, nulls_last=sort_nulls_last)
            if view.columns is not None:
                lf = lf.select(**view.columns)
            if view.limit:
                lf = lf.limit(view.limit)
            return lf.collect()
        else:
            file_path = store.get_table_path(table)
            df = pl.read_parquet(
                str(file_path), n_rows=limit, storage_options=store.get_storage_options("polars", file_path.protocol)
            )
            return df


@ParquetTableStore.register_table(pl)
class LazyPolarsTableHook(sql_hooks.LazyPolarsTableHook):
    pass


@ParquetTableStore.register_table()
class SQLAlchemyTableHook(sql_hooks.SQLAlchemyTableHook):
    @classmethod
    def _create_as_select_statements(
        cls,
        table_name: str,
        schema: Schema,
        query: Select | TextClause | SqlText,
        store: ParquetTableStore,
        suffix: str,
        unlogged: bool,
    ):
        file_path = store.get_table_schema_path(table_name, schema)
        store.metadata_track_view(table_name, schema.get(), file_path.as_uri(), "parquet")
        return [
            CopySelectTo(
                file_path,
                "PARQUET",
                query,
            ),
            CreateViewAsSelect(table_name, schema, store._read_parquet_query(file_path)),
        ]

    @classmethod
    def get_view_query(cls, view: View, store: ParquetTableStore):
        """ParquetTableStore implements special version for src-only views."""
        if view.columns is None and view.sort_by is None and view.limit is None:
            if isinstance(view.src, Iterable):
                try:
                    from sqlalchemy.sql.base import ColumnCollection, ReadOnlyColumnCollection

                    possible = True
                except LookupError:
                    possible = False

                if possible:
                    src_tables = list(view.src)
                    files = [str(store.get_table_path(tbl)) for tbl in src_tables]
                    assert len(src_tables) > 0
                    tbl1 = SQLAlchemyTableHook.retrieve(
                        store, src_tables[0], src_tables[0].stage.current_name, sa.Table
                    )
                    cols = tbl1.c  # this might be oversimplified for categorical columns
                    file_refs = "['" + "','".join(files) + "']"
                    base_from = sa.select(sa.text("*")).select_from(sa.text(f"read_parquet({file_refs})")).alias("sub")

                    # reconstruct columns which were lost by select("*")
                    def bind(c: sa.Column, expr):
                        col = sa.Column(c.name, c.type)
                        col.table = expr
                        return col

                    base_from.c = ReadOnlyColumnCollection(
                        ColumnCollection([(c.name, bind(c, base_from)) for c in cols])
                    )
                    return base_from
        return sql_hooks.get_view_query(view, store)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from tests.fixtures.instances import with_instances

try:
    import sqlalchemy as sa

    # this works only for SQLAlchemy >= 2.0 but is needed for example pipelines
    from sqlalchemy import Alias

    _ = Alias
except ImportError:
    sa = None

try:
    import dataframely as dy
except ImportError:
    dy = None

try:
    import pydiverse.transform as pdt
except ImportError:
    pdt = None

try:
    import pydiverse.colspec as cs
except ImportError:
    cs = None

try:
    import xgboost as xgb
except ImportError:
    xgb = None

examples = []
s3_examples = []
mssql_examples = []
ibmdb_examples = []

if sa:
    from example.run_pipeline import main as example_main
    from example.simple_pipeline import main as simple_pipeline_main
    from example.visualization import main as visualization_main
    from example_imperative.run_pipeline import main as example_imperative_main
    from example_interactive.run_tasks_interactively import main as example_interactive_main
    from example_postgres.run_pipeline_small import main as example_postgres_small

    examples = [
        example_main,
        simple_pipeline_main,
        visualization_main,
        example_imperative_main,
        example_interactive_main,
        example_postgres_small,
    ]

if sa and dy and cs and pdt:
    from example_ibm_db2.run_pipeline import main as example_ibmdb_main
    from example_mssql.run_pipeline import main as example_mssql_main
    from example_parquet_s3.run_pipeline import main as example_s3_main
    from example_postgres.run_pipeline import main as example_postgres_main

    examples.append(example_postgres_main)
    s3_examples.append(example_s3_main)
    mssql_examples.append(example_mssql_main)
    ibmdb_examples.append(example_ibmdb_main)

if sa and cs and pdt and xgb:
    from example_mssql.realistic_pipeline import main as example_mssql_realistic
    from example_parquet_s3.realistic_pipeline import main as example_s3_realistic
    from example_postgres.realistic_pipeline import main as example_postgres_realistic

    examples.append(example_postgres_realistic)
    s3_examples.append(example_s3_realistic)
    mssql_examples.append(example_mssql_realistic)


@pytest.mark.parametrize(
    "fn",
    examples,
)
@with_instances("postgres")
def test_examples_postgres(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()


@pytest.mark.parametrize(
    "fn",
    s3_examples,
)
@with_instances("parquet_s3_backend")
def test_examples_s3(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()


@pytest.mark.parametrize(
    "fn",
    mssql_examples,
)
@with_instances("mssql")
def test_examples_mssql(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()


@pytest.mark.parametrize(
    "fn",
    ibmdb_examples,
)
@with_instances("ibm_db2")
def test_examples_ibmdb(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

"""The imports in this modules are optional dependencies of pipedag or dependencies with multi-version support.

It is easiest in the pipedag codebase to optional dependencies them from this module as opposed to catching
ImportErrors across the codebase
"""



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import contextlib

import pytest

from pydiverse.pipedag import ConfigContext


@contextlib.contextmanager
def swallowing_raises(*args, **kwargs):
    """Swallow logger.error messages and expect an exception to be raised."""
    # Attention: The option name `swallow_exceptions` is a bit misleading. It does not
    #   swallow the exception, but it swallows the error message that occurs between
    #   catching and rethrowing the exception and turns it into an info message.
    with ConfigContext.get().evolve(swallow_exceptions=True):
        with pytest.raises(*args, **kwargs) as raises:
            yield raises



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from packaging.version import Version

try:
    import prefect

    prefect_version = Version(prefect.__version__)
except ImportError:
    prefect = None
    prefect_version = Version("0")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import structlog

from pydiverse.pipedag.optional_dependency.boto import ClientError, boto3


def create_bucket_if_not_exists(bucket_name: str, s3_client, region: str | None = None) -> bool:
    logger = structlog.get_logger(__name__ + ":create_bucket_if_not_exists")
    try:
        s3_client.head_bucket(Bucket=bucket_name)
        logger.info(f"Bucket '{bucket_name}' already exists.")
        return False
    except ClientError as e:
        error_code = int(e.response["Error"]["Code"])
        if error_code == 404:
            # Bucket does not exist, safe to create
            s3_client.create_bucket(
                Bucket=bucket_name,
                CreateBucketConfiguration={"LocationConstraint": region} if region is not None else {},
            )
            logger.info(f"Bucket '{bucket_name}' created.")
            return True
        elif error_code == 403:
            logger.info(f"Bucket '{bucket_name}' exists but is owned by someone else.")
            return False
        else:
            raise


def initialize_test_s3_bucket(
    host: str = "127.0.0.1", port: str | int = 9000, test_bucket: str = "pipedag-test-bucket"
) -> bool:
    if boto3 is not None:
        minio_client = boto3.client(
            "s3",
            endpoint_url=f"http://{host}:{port}",
            aws_access_key_id="minioadmin",  # default MinIO credentials
            aws_secret_access_key="minioadmin",
            region_name="us-east-1",
            config=boto3.session.Config(s3={"addressing_style": "path"}),
        )
        return create_bucket_if_not_exists(test_bucket, minio_client)
    return False



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import types

try:
    # pydot should be an optional dependency
    import pydot
except ImportError:
    pydot = types.ModuleType("pydot")
    pydot.Dot = None



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# pipedag supports both sqlalchemy < 2 and >= 2

try:
    from sqlalchemy import URL, Alias, Connection, Engine, Select, TextClause
    from sqlalchemy import Text as SqlText
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.engine import URL, Connection, Engine
    from sqlalchemy.sql.expression import Alias, Select, TextClause

    SqlText = TextClause  # this is what sa.text() returns


__all__ = [
    "URL",
    "Alias",
    "Connection",
    "Engine",
    "Select",
    "SqlText",
    "TextClause",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy

import pandas as pd
import pytest
import sqlalchemy as sa
import structlog
from pandas.testing import assert_frame_equal

from pydiverse.pipedag import ConfigContext, Flow, Stage, StageLockContext, Table
from pydiverse.pipedag.container import Schema
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.core.config import PipedagConfig, create_basic_pipedag_config
from tests.fixtures.instances import with_instances
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2


@with_instances(
    "postgres",
    "local_table_cache",
    "local_table_cache_inout",
    "local_table_cache_inout_numpy",
)
@pytest.mark.lock_tests  # this test seems to hang easily in ColSpec tests
@pytest.mark.parametrize("imperative", [False, True], ids=["imperative", "not_imperative"])
@pytest.mark.parametrize(
    "write_local_table_cache", [True, False], ids=["write_local_table_cache", "no_write_local_table_cache"]
)
def test_get_output_from_store(mocker, imperative, write_local_table_cache):
    _m = m2 if imperative else m
    with Flow() as f:
        with Stage("stage_1") as s:
            df1 = _m.pd_dataframe({"x": [0, 1, 2, 3]})
            df2 = _m.pd_dataframe({"y": [0, 1, 2, 3]})
            dataframes = _m.create_tuple(df1, df2)

    # We only use the StageLockContext for testing
    with StageLockContext():
        # First run to ensure cache-validity -> No local cache writes during the second flow run
        f.run()
        local_table_cache = ConfigContext.get().store.local_table_cache
        # A local table cache only exists when configured
        if local_table_cache:
            local_table_cache.clear_cache(s)
            store_input_spy = mocker.spy(local_table_cache, "store_input")
            _store_table_spy = mocker.spy(local_table_cache, "_store_table")
            retrieve_table_obj_spy = mocker.spy(local_table_cache, "retrieve_table_obj")
            _retrieve_table_obj_spy = mocker.spy(local_table_cache, "_retrieve_table_obj")
        result = f.run()

        # Call on MaterializingTask
        pd.testing.assert_frame_equal(
            df1.get_output_from_store(as_type=pd.DataFrame, write_local_table_cache=write_local_table_cache),
            result.get(df1, as_type=pd.DataFrame),
        )

        pd.testing.assert_frame_equal(
            df2.get_output_from_store(as_type=pd.DataFrame, write_local_table_cache=write_local_table_cache),
            result.get(df2, as_type=pd.DataFrame),
        )

        pd.testing.assert_frame_equal(
            dataframes.get_output_from_store(as_type=pd.DataFrame, write_local_table_cache=write_local_table_cache)[1],
            result.get(dataframes, as_type=pd.DataFrame)[1],
        )

        # since df1 and df2 are outputs from tasks with the same name (pd_dataframe)
        # ignoring the position hash when retrieving
        # should just return the output from the latest call
        pd.testing.assert_frame_equal(
            df1.get_output_from_store(
                as_type=pd.DataFrame, ignore_position_hashes=True, write_local_table_cache=write_local_table_cache
            ),
            result.get(df2, as_type=pd.DataFrame),
        )

        # Call on MaterializingTaskGetItem
        pd.testing.assert_frame_equal(
            dataframes[0].get_output_from_store(as_type=pd.DataFrame, write_local_table_cache=write_local_table_cache),
            dataframes.get_output_from_store(as_type=pd.DataFrame, write_local_table_cache=write_local_table_cache)[0],
        )

        pd.testing.assert_frame_equal(
            dataframes[0].get_output_from_store(
                as_type=pd.DataFrame, ignore_position_hashes=True, write_local_table_cache=write_local_table_cache
            ),
            dataframes.get_output_from_store(
                as_type=pd.DataFrame, ignore_position_hashes=True, write_local_table_cache=write_local_table_cache
            )[0],
        )

        if local_table_cache:
            # We try to get a table 16 times after running the flow.
            expected_retrieve_table_obj = 16
            # If write_local_table_cache we attempt to store each of df1, df2, dataframes[0], dataframes[1]
            # once because it was not yet in the cache. Otherwise, storing is never successful, so we attempt it for
            # each attempted retrieval.
            expected_store_input = 4 if write_local_table_cache else expected_retrieve_table_obj
            expected_successful_retrieve_table_obj = expected_retrieve_table_obj - expected_store_input

            assert store_input_spy.call_count == expected_store_input
            assert _store_table_spy.call_count == expected_store_input
            assert retrieve_table_obj_spy.call_count == expected_retrieve_table_obj
            assert _retrieve_table_obj_spy.call_count == expected_successful_retrieve_table_obj


@with_instances("postgres")
@pytest.mark.parametrize("imperative", [False, True])
def test_call_task_outside_flow(imperative):
    _m = m2 if imperative else m
    # Literal Values
    assert m.one() == 1
    assert m.create_tuple(m.one(), m.two()) == (1, 2)
    assert m.noop("foo") == "foo"

    # Table
    df1 = _m.pd_dataframe({"x": [0, 1, 2, 3]})
    df2 = m.noop(df1)
    expected = pd.DataFrame({"x": [0, 1, 2, 3]})

    for df in [df1, df2]:
        pd.testing.assert_frame_equal(df, expected)
        _m.assert_table_equal(df, expected)

    # Lazy Table
    expr = _m.simple_lazy_table()
    expected_expr = (
        "SELECT 0 AS col1, '0' AS col2 "
        "UNION ALL SELECT 1 AS col1, '1' AS col2 "
        "UNION ALL SELECT 2 AS col1, '2' AS col2 "
        "UNION ALL SELECT 3 AS col1, '3' AS col2"
    )
    config_context = ConfigContext.get()
    assert (
        str(
            expr.compile(
                config_context.store.table_store.engine,
                compile_kwargs={"literal_binds": True},
            )
        )
        == expected_expr
    )

    tbl = m.noop(sa.Table("foo", sa.MetaData()).alias("bar"))
    assert tbl.name == "bar"
    assert tbl.original.name == "foo"

    # Blob
    expected_blob = {"a": {"b": 1, "c": 2}, "d": [1, (2, 3)]}
    blob = m.as_blob(expected_blob)

    assert blob is expected_blob
    m.assert_blob_equal(blob, expected_blob)

    # RawSql
    expected_query = "SELECT 1 as x"
    raw_sql = m.as_raw_sql(expected_query)

    assert raw_sql == expected_query


@with_instances("postgres")
def test_imperative_materialize_given_config():
    _m = m2
    df = pd.DataFrame({"x": [0, 1, 2, 3]})
    tbl = Table(df)
    assert tbl.materialize() is df  # should be no-op
    assert tbl.materialize(return_nothing=True) is None

    config_context = ConfigContext.get()
    engine = config_context.store.table_store.engine
    with engine.connect() as conn:
        conn.execute(sa.text("DROP SCHEMA IF EXISTS dummy_schema CASCADE"))
        conn.execute(sa.text("CREATE SCHEMA IF NOT EXISTS dummy_schema"))
        if sa.__version__ >= "2.0.0":
            conn.commit()

    # test failure on materialization without schema
    with pytest.raises(ValueError, match="schema must be provided"):
        tbl.materialize(config_context)
    # test materialization with explicit config_context and direct schema
    ref = tbl.materialize(config_context, schema=Schema("dummy_schema"))
    sa_tbl = sa.Table(ref.original.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    sa_tbl2 = _m.noop_lazy(sa_tbl.alias("alias"))
    assert sa_tbl2.name == "alias"
    assert sa_tbl2.original.name == sa_tbl.name

    # test return nothing materialization
    tbl.name = "a"
    assert tbl.materialize(config_context, schema=Schema("dummy_schema"), return_nothing=True) is None
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and prefixed schema
    tbl.name = "b"
    tbl.materialize(config_context, schema=Schema("schema", prefix="dummy_"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and postfixed schema
    tbl.name = "c"
    tbl.materialize(config_context, schema=Schema("dummy", suffix="_schema"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context incl. prefixed schema
    pipedag_config = create_basic_pipedag_config(
        config_context.store.table_store.engine.url.render_as_string(hide_password=False)
    )
    raw_config2 = copy.deepcopy(pipedag_config.raw_config)
    raw_config2["instances"]["__any__"]["table_store"]["args"]["schema_prefix"] = "dummy_"
    cfg2 = PipedagConfig(raw_config2).get()
    tbl.name = "e"
    with pytest.raises(ValueError, match="Schema prefix and postfix must match"):
        tbl.materialize(cfg2, schema=Schema("schema", prefix="dummyX_"))
    tbl.materialize(cfg2, schema=Schema("schema", prefix="dummy_"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name

    # test materialization with explicit config_context and postfixed schema
    pipedag_config = create_basic_pipedag_config(
        config_context.store.table_store.engine.url.render_as_string(hide_password=False)
    )
    raw_config2 = copy.deepcopy(pipedag_config.raw_config)
    raw_config2["instances"]["__any__"]["table_store"]["args"]["schema_suffix"] = "_schema"
    cfg2 = PipedagConfig(raw_config2).get()
    tbl.name = "e"
    with pytest.raises(ValueError, match="Schema prefix and postfix must match"):
        tbl.materialize(cfg2, schema=Schema("dummy", suffix="_Xschema"))
    tbl.materialize(cfg2, schema=Schema("dummy", suffix="_schema"))
    sa_tbl = sa.Table(tbl.name, sa.MetaData(), schema="dummy_schema", autoload_with=engine)
    assert [c.name for c in sa_tbl.columns] == ["x"]
    assert sa_tbl.name == tbl.name


@with_instances("parquet_s3_backend")
def test_metadata_store_synchronization():
    logger = structlog.get_logger(__name__ + ".test_metadata_store_synchronization")
    x = {"x": [0, 1, 2, 3]}
    y = {"y": [0, 1, 2, 3]}
    constant = dict(c=[0])
    with Flow() as f:
        with Stage("stage_1") as s:
            df1 = m.pd_dataframe(x)
            df2 = m.pd_dataframe(y)
            dataframes = m.create_tuple(df1, df2)
            out = m.noop(constant)  # prevent 100% cache valid stage
            out2 = m.pd_dataframe(out)
            _ = dataframes, out2

    cfg1 = ConfigContext.get()
    # setup second configuration with different local duckdb file
    cfg2 = cfg1.evolve(
        _config_dict=dict(
            table_store=dict(args=dict(url="duckdb:////tmp/pipedag/parquet_duckdb/parquet_s3_backend2.duckdb"))
        ),
        _transfer_cache_=False,
    )

    is_odd = {}
    for mode in [CacheValidationMode.FORCE_CACHE_INVALID, CacheValidationMode.NORMAL]:
        for cfg in (cfg1, cfg2):
            store = cfg.store.table_store  # keep this before f.run() to allow transfer of cached_property store
            logger.info("** Running Pipeline for **", mode=mode, db_file=store.engine.url)
            with StageLockContext():
                f.run(config=cfg, cache_validation_mode=mode)
                out_x, out_y = dataframes.get_output_from_store(as_type=pd.DataFrame)
                assert_frame_equal(out_x, pd.DataFrame(x))
                assert_frame_equal(out_y, pd.DataFrame(y))
                query = sa.text(f"""
                    FROM duckdb_views() SELECT sql WHERE
                        schema_name='{store.get_schema(s.name).get()}'
                """)
                with store.engine.connect() as conn:
                    views = store.execute(query, conn=conn).fetchall()
                cnt_odd = sum(1 if "FROM stage_1__odd." in v[0] else 0 for v in views)
                cnt_even = sum(1 if "FROM stage_1__even." in v[0] else 0 for v in views)
                assert len(views) == 5
                assert cnt_odd + cnt_even == 5
                assert min(cnt_odd, cnt_even) == 0
                is_odd[id(cfg)] = cnt_odd > 0

            if mode == CacheValidationMode.NORMAL:
                constant["c"][0] += 1  # prevent 100% cache valid stage

        assert is_odd[id(cfg1)] != is_odd[id(cfg2)]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError:
    boto3 = None
    ClientError = None



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import copy
import getpass
import os
import time
from contextlib import contextmanager
from typing import Literal

import pandas as pd
import polars as pl
import sqlalchemy as sa

from pydiverse.pipedag import Schema, Table
from pydiverse.pipedag.backend.table.sql import hooks
from pydiverse.pipedag.backend.table.sql.ddl import AddIndex
from pydiverse.pipedag.backend.table.sql.hooks import (
    IbisTableHook,
)
from pydiverse.pipedag.backend.table.sql.sql import SQLTableStore
from pydiverse.pipedag.optional_dependency.ibis import ibis
from pydiverse.pipedag.optional_dependency.snowflake import snowflake


class SnowflakeTableStore(SQLTableStore):
    """
    SQLTableStore that supports `Snowflake`_.

    Takes the same arguments as
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`
    """

    _dialect_name = "snowflake"
    # could be made configurable in pipedag.yaml if needed:
    ADBC_MAX_RETRIES = 3
    SQLALCHEMY_PRE_PING = False
    SQLALCHEMY_RECYCLE = 1200
    SQLALCHEMY_LIFO = True

    def _create_engine(self, **kwargs):
        connect_args = dict(kwargs.pop("connect_args", {}) or {})
        connect_args.setdefault("client_session_keep_alive", True)
        session_parameters = dict(connect_args.get("session_parameters", {}) or {})
        session_parameters.setdefault("QUERY_TAG", self.get_adbc_query_tag("|sqa"))
        connect_args["session_parameters"] = session_parameters

        # knobs you can expose/configure:
        kwargs.setdefault("pool_pre_ping", self.SQLALCHEMY_PRE_PING)
        kwargs.setdefault("pool_recycle", self.SQLALCHEMY_RECYCLE)
        kwargs.setdefault("pool_use_lifo", self.SQLALCHEMY_LIFO)
        kwargs["connect_args"] = connect_args

        return super()._create_engine(**kwargs)

    def _default_isolation_level(self) -> str | None:
        return None  # "READ UNCOMMITTED" does not exist in Snowflake

    def optional_pause_for_db_transactionality(
        self,
        prev_action: Literal[
            "table_drop",
            "table_create",
            "schema_drop",
            "schema_create",
            "schema_rename",
        ],
    ):
        _ = prev_action
        # The snowflake backend has transactionality problems with very quick
        # DROP/CREATE or RENAME activities for both schemas and tables
        # which happen in testing.
        time.sleep(0)  # trying with 0 again. This was 2s at some point

    def _init_database(self):
        create_database = self.engine_url.database.split("/")[0]
        with self.engine.connect() as conn:
            if not [
                x.name
                for x in conn.exec_driver_sql("SHOW DATABASES").mappings().all()
                if x.name.upper() == create_database.upper()
            ]:
                self._init_database_with_database(
                    "snowflake",
                    disable_exists_check=True,
                    create_database=create_database,
                )

    # --- connection open/apply-session -----------------------------------------
    def _adbc_open(self):
        import adbc_driver_snowflake.dbapi as adbc

        engine = self.engine
        adbc_query_tag = self.get_adbc_query_tag()
        raw_kwargs = {
            "adbc.snowflake.sql.account": engine.url.host,
            "adbc.snowflake.sql.db": engine.url.database.split("/")[0] if engine.url.database else None,
            "adbc.snowflake.sql.schema": (
                engine.url.database.split("/")[1] if engine.url.database and "/" in engine.url.database else None
            ),
            "adbc.snowflake.sql.warehouse": getattr(self, "warehouse", None) or engine.url.query.get("warehouse"),
            "adbc.snowflake.sql.role": getattr(self, "role", None) or engine.url.query.get("role"),
            "adbc.snowflake.sql.auth_type": "auth_pat",
            "adbc.snowflake.sql.client_option.auth_token": engine.url.password,
            "adbc.snowflake.sql.client_option.use_high_precision": "false",
            "adbc.snowflake.sql.client_option.keep_session_alive": "true",
        }
        db_kwargs = {k: v for k, v in raw_kwargs.items() if v is not None}
        conn = adbc.connect(db_kwargs=db_kwargs)
        with conn.cursor() as cur:
            cur.execute(f"ALTER SESSION SET QUERY_TAG='{adbc_query_tag}'")
            cur.execute("ALTER SESSION SET CLIENT_SESSION_KEEP_ALIVE=TRUE")
            cur.execute("ALTER SESSION SET CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY=900")  # 15 min
            cur.execute("ALTER SESSION SET QUOTED_IDENTIFIERS_IGNORE_CASE=TRUE")  # ADBC append cannot do unquoted names
        self.keep_alive_adbc_connection = conn
        return conn

    def get_adbc_query_tag(self, suffix="") -> str:
        adbc_query_tag = f"pipedag|{self.instance_id}|user:{getpass.getuser()}|pid:{os.getpid()}{suffix}"
        return adbc_query_tag

    def adbc_connect(self):
        if self.keep_alive_db_connections is None:
            return self._adbc_open()
        return self.keep_alive_db_connections

    # --- retry API ---
    def adbc_reconnect(self, state: list) -> bool:
        """
        Gate for the caller's while-loop. Initializes state and returns
        True while we still have attempts left (including the first try).

        Example usage:
        state = []
        while store.adbc_reconnect(state):
            with store.adbc_reconnect_attempt(state) as conn:
                pl.read_database(query, connection=conn)
        """
        if len(state) != 3:
            # [attempt_idx, max_retries, last_exc]
            state[:] = [0, self.ADBC_MAX_RETRIES, None]
            return True

        attempt, max_retries, last_exc = state
        if attempt <= max_retries:
            return last_exc is not None

        # exhausted → re-raise the last captured OperationalError (if any)
        if last_exc is not None:
            raise last_exc
        raise RuntimeError("ADBC retries exhausted with no captured exception")

    @contextmanager
    def adbc_reconnect_attempt(self, state: list):
        """
        Yields a live ADBC connection.
        On OperationalError:
          - capture the exception in state[2]
          - increment attempt counter in state[0]
          - reopen sticky connection so next loop has a fresh one
          - suppress the error so the loop can iterate
        """
        assert len(state) == 3
        import adbc_driver_snowflake.dbapi as adbc

        conn = self.adbc_connect()
        try:
            yield conn
        except adbc.OperationalError as exc:
            # store exception
            state[2] = exc
            # bump attempt index
            state[0] += 1
            # teardown & reopen the sticky connection so next loop has a fresh one
            try:
                if self.keep_alive_adbc_connection is not None:
                    self.keep_alive_adbc_connection.close()
            finally:
                self.keep_alive_adbc_connection = None
                self._adbc_open()
            # swallow the error so the while-loop continues
        # let all other exceptions propagate

    def add_index(
        self,
        table_name: str,
        schema: Schema,
        index_columns: list[str],
        name: str | None = None,
    ):
        self.logger.warning(
            "Snowflake does not support creating indexes on snowflake tables",
            query_not_executed=str(AddIndex(table_name, schema, index_columns, name)),
        )

    def reflect_table(self, table_name: str, schema: str | Schema) -> sa.Table:
        if isinstance(schema, Schema):
            schema = schema.get()
        if schema.count(".") == 1:
            # sqlalchemy snowflake driver does not support schema="database.schema"
            database, schema_part = schema.split(".")
            with self.engine.connect() as conn:
                conn.execute(sa.text(f"USE DATABASE {database}"))
                tbl = sa.Table(
                    table_name,
                    sa.MetaData(),
                    schema=schema_part,
                    autoload_with=conn,
                )
                tbl.schema = schema  # restore full schema name
                conn.execute(sa.text(f"USE DATABASE {self.engine.url.database}"))
        else:
            tbl = sa.Table(
                table_name,
                sa.MetaData(),
                schema=schema,
                autoload_with=self.engine,
            )
        return tbl

    def has_table_or_view(self, name: str, schema: Schema | str):
        if isinstance(schema, Schema):
            schema = schema.get()
        if schema.count(".") == 1:
            # sqlalchemy snowflake driver does not support schema="database.schema"
            database, schema_part = schema.split(".")
            with self.engine.connect() as conn:
                conn.execute(sa.text(f"USE DATABASE {database}"))
                has_table = self._has_table_or_view(conn, name, schema_part)
                conn.execute(sa.text(f"USE DATABASE {self.engine.url.database}"))
        else:
            has_table = self._has_table_or_view(self.engine, name, schema)
        return has_table

    @staticmethod
    def _has_table_or_view(conn, name: str, schema: str) -> bool:
        inspector = sa.inspect(conn)
        has_table = inspector.has_table(name, schema=schema)
        # workaround for sqlalchemy backends that fail to find views with has_table
        # in particular DuckDB https://github.com/duckdb/duckdb/issues/10322
        if not has_table:
            has_table = name in inspector.get_view_names(schema=schema)
        return has_table


@SnowflakeTableStore.register_table(snowflake)
class SQLAlchemyTableHook(hooks.SQLAlchemyTableHook):
    @classmethod
    def retrieve(
        cls,
        store: SQLTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type[sa.Table],
        limit: int | None = None,
    ) -> sa.sql.expression.Selectable:
        tbl = cls.__base__.retrieve(store, table, stage_name, as_type, limit)

        # fix integer types
        def bind(c: sa.Column, expr):
            col = sa.Column(c.name, c.type)
            col.table = expr
            return col

        def fix(c: sa.Column) -> sa.Column:
            # convert DECIMAL(38,0) to BIGINTEGER
            if isinstance(c.type, sa.DECIMAL) and c.type.precision == 38 and c.type.scale == 0:
                new_c = copy.deepcopy(c)
                new_c.type = sa.BigInteger()
                return new_c
            return c

        try:
            from sqlalchemy.sql.base import ColumnCollection, ReadOnlyColumnCollection

            tbl.c = ReadOnlyColumnCollection(ColumnCollection([(c.name, bind(fix(c), tbl)) for c in tbl.c]))
        except ImportError:
            store.logger.info("Can't fix snowflake sqlalchemy types for sqlalchemy < 2", columns=list(tbl.c))
        return tbl


@SnowflakeTableStore.register_table(snowflake)
class PandasTableHook(hooks.PandasTableHook):
    @classmethod
    def upload_table(
        cls,
        table: Table[pd.DataFrame],
        schema: str,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        df = table.obj
        # snowflake driver cannot receive object column with datetime objects
        obj_cols = df.dtypes[df.dtypes == object]  # noqa:E721
        if len(obj_cols) > 0:
            datetime_cols = []
            datetime_ns_cols = []

            def compatible(series: pd.Series, _type: str):
                return (series[~series.isna()].astype(_type) == series[~series.isna()]).all()

            for col in obj_cols.index:
                # detect datetime columns
                try:
                    if compatible(df[col], "datetime64[us]"):
                        datetime_cols.append(col)
                    elif compatible(df[col], "datetime64[ns]"):
                        datetime_ns_cols.append(col)
                except ValueError:
                    pass
            df[datetime_cols] = df[datetime_cols].astype("datetime64[us]")
            df[datetime_ns_cols] = df[datetime_ns_cols].astype("datetime64[ns]")
        super().upload_table(table, schema, dtypes, store, early)


@SnowflakeTableStore.register_table(snowflake)
class PolarsTableHook(hooks.PolarsTableHook):
    @classmethod
    def dialect_supports_connectorx(cls):
        # ConnectorX (used by Polars read_database_uri) does not support Snowflake.
        return False

    @classmethod
    def dialect_wrong_polars_column_names(cls):
        # for Snowflake, polars returns uppercase column names by default
        return True

    @classmethod
    def adbc_write_database(
        cls,
        df: pl.DataFrame,
        store: SQLTableStore,
        schema_name: str,
        table_name: str,
        if_table_exists: Literal["replace", "append", "fail"] = "append",
    ) -> int:
        def _quote_ident(name: str) -> str:
            assert '"' not in name, f"Let's not test escaping rules for snowflake with quote in: '{name}'"
            return '"' + name + '"'  # we set session to case-insensitive identifiers above (ADBC cannot do unquoted)

        state: list = []
        while store.adbc_reconnect(state):
            with store.adbc_reconnect_attempt(state) as conn:
                # workaround ADBC/snowflake bug that it cannot transport schema:
                with conn.cursor() as cur:
                    cur.execute(f"USE SCHEMA {_quote_ident(schema_name)}")

                return df.write_database(
                    _quote_ident(table_name),  # schema see above
                    connection=conn,
                    if_table_exists=if_table_exists,
                    engine="adbc",
                )

    @classmethod
    def adbc_read_database(cls, query, store: SQLTableStore) -> pl.DataFrame:
        state: list = []
        while store.adbc_reconnect(state):
            with store.adbc_reconnect_attempt(state) as conn:
                df = pl.read_database(query, connection=conn)
                # TODO: check whether this is still needed
                # # convert Decimal(38,0) to Int64 (snowflake treats them interchangeably)
                # df = df.with_columns(pl.selectors.by_dtype(pl.Decimal(38, 0)).cast(pl.Int64))
                return df


@SnowflakeTableStore.register_table(ibis.api.Table)
class IbisTableHook(IbisTableHook):
    @classmethod
    def _conn(cls, store: SnowflakeTableStore):
        return ibis.snowflake._from_url(store.engine_url)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from concurrent.futures.process import BrokenProcessPool

import pandas as pd
import pytest
import sqlalchemy as sa
import structlog

from pydiverse.pipedag import AUTO_VERSION, ConfigContext, Flow, Stage, Table, materialize
from pydiverse.pipedag.container import SortCol, SortOrder, View
from pydiverse.pipedag.context import FinalTaskState, RunContext, StageLockContext
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.context.trace_hook import PrintTraceHook
from pydiverse.pipedag.core.config import PipedagConfig
from pydiverse.pipedag.optional_dependency.sqlalchemy import Alias
from pydiverse.pipedag.optional_dependency.transform import C, pdt

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import (
    ALL_INSTANCES,
    ORCHESTRATION_INSTANCES,
    skip_instances,
    with_instances,
)
from tests.util import select_as, swallowing_raises
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2

pytestmark = [with_instances(tuple(set(ALL_INSTANCES) - {"snowflake"}), ORCHESTRATION_INSTANCES)]


def test_materialize_literals():
    """
    Check that all json serializable literals get stored and retrieved
    correctly from the table / metadata store.
    """

    def materialize_and_retrieve(value):
        v = m.noop(value)
        return m.assert_equal(value, v)

    with Flow("flow") as f:
        with Stage("stage"):
            materialize_and_retrieve(None)

            materialize_and_retrieve(0)
            materialize_and_retrieve(1)

            # max 32bit
            materialize_and_retrieve(-2147483648)
            materialize_and_retrieve(2147483647)

            # max 64bit
            materialize_and_retrieve(-9223372036854775808)
            materialize_and_retrieve(9223372036854775807)

            # in python json we even have infinite precision integers:
            materialize_and_retrieve(-9223372036854775809)
            materialize_and_retrieve(9223372036854775808)

            materialize_and_retrieve(3.14)
            materialize_and_retrieve(-2.71)

            materialize_and_retrieve("")
            materialize_and_retrieve("a string")
            materialize_and_retrieve("a" * 256)

            materialize_and_retrieve([0, 1, 2, 3])

            materialize_and_retrieve({"0": 0, "1": 1})

            materialize_and_retrieve(
                {
                    "key": [0, 1, 2, 3],
                    "numbers": {
                        "zero": 0,
                        "one": 1,
                    },
                    "nested_lists": [[[None]]],
                }
            )

    assert f.run().successful


@with_instances("snowflake")
@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe()
            _m.assert_table_equal(x, x)

    assert f.run().successful


@materialize(input_type=sa.Table, lazy=True)
def create_view1(tbl: Alias):
    return Table(View(src=tbl, sort_by=tbl.c.col2, columns=tbl.c.col1.label("c1"), limit=2))


@materialize(input_type=sa.Table, lazy=True)
def create_view2(tbl: Alias, tbl2: Alias):
    return Table(View(src=[tbl, tbl2], sort_by=[tbl.c.col2.desc(), tbl.c.col1], columns=[tbl.c.col1], limit=2))


@materialize(input_type=sa.Table, lazy=True)
def create_view_view1(tbl: Alias):
    return Table(View(src=tbl, columns=tbl.c.c1))


@materialize(input_type=sa.Table, lazy=True)
def create_view_view2(tbl: Alias):
    return Table(View(src=tbl, sort_by=SortCol(tbl.c.col1, SortOrder.DESC), limit=3))


@materialize(input_type=pdt.SqlAlchemy, lazy=True)
def create_view_pdt1(tbl: pdt.Table):
    return Table(View(src=tbl, sort_by=tbl.col2, columns=dict(c1=tbl.col1), limit=2))


@materialize(input_type=pdt.SqlAlchemy, lazy=True)
def create_view_pdt2(tbl: pdt.Table, tbl2: pdt.Table):
    return Table(View(src=[tbl, tbl2], sort_by=[tbl.col2.descending(), tbl2.col1], columns=[C.col1], limit=2))


@materialize(input_type=pdt.SqlAlchemy, lazy=True)
def create_view_pdt2b(tbl: pdt.Table, tbl2: pdt.Table):
    # currently, multi-source views are not supported by consuming tasks with input_type=pdt.Sqlalchemy
    return Table(View(src=tbl, sort_by=[tbl.col2.descending(), tbl2.col1], columns=[C.col1], limit=2))


@materialize(input_type=pdt.SqlAlchemy, lazy=True)
def create_view_view_pdt1(tbl: pdt.Table):
    return Table(View(src=tbl, columns=tbl.c1))


@materialize(input_type=pdt.SqlAlchemy, lazy=True)
def create_view_view_pdt2(tbl: pdt.Table):
    return Table(View(src=tbl, sort_by=[tbl.col1], limit=3))


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_view(imperative):
    logger = structlog.getLogger(__name__ + ".test_materialize_view")
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe()
            x2 = _m.noop(x)
            y = create_view1(x)
            z = create_view2(x, x2)
            k = create_view_view1(y)
            n = create_view_view2(z)
            yy = _m.noop(y)
            zz = _m.noop(z)
            kk = _m.noop(k)
            nn = _m.noop(n)
            yy_lazy = _m.noop_lazy(y)
            zz_lazy = _m.noop_lazy(z)
            kk_lazy = _m.noop_lazy(k)
            nn_lazy = _m.noop_lazy(n)
            _ = yy, zz, kk, nn, yy_lazy, zz_lazy, kk_lazy, nn_lazy

    assert f.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID).successful

    for i in range(2):
        logger.info(f"** Starting run: {i}")
        # run three times to test some caching behavior (this is at least a smoke test for caching)
        assert f.run().successful


@pytest.mark.skipif(C is None, reason="Pydiverse.transform not available")
@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_view_union(imperative):
    """The implementation of this special case is different in ParquetTableStore."""
    logger = structlog.getLogger(__name__ + ".test_materialize_view")

    @materialize(input_type=sa.Table, lazy=True)
    def create_view(tbl: Alias, tbl2: Alias):
        return Table(View(src=[tbl, tbl2]))

    @materialize(input_type=pdt.SqlAlchemy, lazy=True)
    def create_view_pdt(tbl: pdt.Table, tbl2: pdt.Table):
        return Table(View(src=[tbl, tbl2]))

    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe()
            x2 = _m.noop(x)
            y = create_view(x, x2)
            z = create_view_pdt(x, x2)
            yy_lazy = _m.noop_lazy(y)
            zz_lazy = _m.noop_lazy(z)
            yy = _m.noop(y)
            zz = _m.noop(z)
            _ = yy, zz, yy_lazy, zz_lazy

    assert f.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID).successful

    for i in range(2):
        logger.info(f"** Starting run: {i}")
        # run three times to test some caching behavior (this is at least a smoke test for caching)
        assert f.run().successful


@pytest.mark.skipif(C is None, reason="Pydiverse.transform not available")
@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_view_pdt(imperative):
    @materialize(input_type=pdt.SqlAlchemy, lazy=True)
    def noop_lazy_pdt(tbl: pdt.Table, name: str | None = None):
        if name is None:
            return tbl
        return tbl >> pdt.alias(name)

    @materialize(input_type=pdt.Polars, version=AUTO_VERSION)
    def noop_pdt(tbl: pdt.Table, name: str | None = None):
        if name is None:
            return tbl
        return tbl >> pdt.alias(name)

    logger = structlog.getLogger(__name__ + ".test_materialize_view_pdt")
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe()
            x2 = _m.noop(x)
            a = create_view1(x)
            y = create_view_pdt1(x)
            z = create_view_pdt2(x, x2)
            zb = create_view_pdt2b(x, x2)  # workaround
            k = create_view_view_pdt1(y)
            la = create_view_view_pdt1(a)
            n = create_view_view_pdt2(zb)
            yy = _m.noop(y)
            zz = _m.noop(z)
            kk = _m.noop(k)
            nn = _m.noop(n)
            pdt_aa = noop_pdt(a, name="aa")
            pdt_yy = noop_pdt(y, name="yy")
            pdt_zz = noop_pdt(z, name="zz")
            pdt_kk = noop_pdt(k, name="kk")
            pdt_nn = noop_pdt(n, name="nn")
            yy_lazy = _m.noop_lazy(y)
            zz_lazy = _m.noop_lazy(z)
            pdt_aa_lazy = noop_lazy_pdt(a, name="laa")
            pdt_yy_lazy = noop_lazy_pdt(y, name="lyy")
            pdt_zz_lazy = noop_lazy_pdt(zb, name="lzz")
            pdt_kk_lazy = noop_lazy_pdt(k, name="lkk")
            pdt_nn_lazy = noop_lazy_pdt(n, name="lnn")
            _ = (
                la,
                yy,
                zz,
                kk,
                nn,
                pdt_aa,
                pdt_yy,
                pdt_zz,
                pdt_kk,
                pdt_nn,
                yy_lazy,
                zz_lazy,
                pdt_aa_lazy,
                pdt_yy_lazy,
                pdt_zz_lazy,
                pdt_kk_lazy,
                pdt_nn_lazy,
            )

    for i in range(3):
        logger.info(f"** Starting run: {i}")
        # run three times to test some caching behavior (this is at least a smoke test for caching)
        assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table_subtask(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_subtask()
            y = _m.noop_subtask(x)
            z = _m.noop_subtask_lazy(x)
            _m.assert_table_equal(x, y)
            _m.assert_table_equal(x, z)

    assert f.run(cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID).successful
    assert f.run().successful
    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table_subtask_fail_input_type(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_subtask()
            _ = _m.noop_subtask_fail_input_type(x)
    with swallowing_raises(RuntimeError, match="does not match parent task input type"):
        try:
            f.run()
        except BrokenProcessPool as e:
            # TODO: find out whether we can avoid BrokenProcessPool from happening
            raise RuntimeError("does not match parent task input type + BrokenProcessPool") from e


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table_twice(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe()
            y = _m.simple_dataframe()

            m.assert_table_equal(x, y)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_debug_materialize_table_no_taint(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_debug_materialize_no_taint()
            _m.assert_table_equal(x, x)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_debug_materialize_table_twice(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_debug_materialize_twice()
            _m.assert_table_equal(x, x)

    with swallowing_raises(RuntimeError, match="interactive debugging"):
        assert f.run().successful


@with_instances("postgres")
def test_imperative_minimal_example():
    with Flow("flow") as f:
        with Stage("input"):
            in_ = m.simple_dataframe()
        with Stage("stage"):
            _ = m.complex_imperative_materialize(in_)

    trace_hook = PrintTraceHook()  # can be used to debug caching issues
    result1 = f.run(
        cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
        trace_hook=trace_hook,
    )

    result2 = f.run(trace_hook=trace_hook)

    print("Run 1:")
    for task in f.tasks:
        print(f"{task._name}: {result1.task_states.get(task)}")

    print("Run 2:")
    for task in f.tasks:
        print(f"{task._name}: {result2.task_states.get(task)}")

    # Expectation: In Run 2 all tasks should be cached
    for task in f.tasks:
        if not task._name.startswith("Commit"):
            assert result2.task_states.get(task) == FinalTaskState.CACHE_VALID


@with_instances("postgres")
def test_materialize_blob():
    with Flow("flow") as f:
        with Stage("stage_0"):
            x = m.object_blob({"a": 12, "b": 24})
            y = m.as_blob(x)
            m.assert_blob_equal(x, y)

        with Stage("stage_1"):
            one = m.one()
            one_blob = m.as_blob(m.one())
            m.assert_equal(one, one_blob)

    assert f.run().successful


def test_failure():
    with Flow("flow") as f:
        with Stage("failure_stage"):
            m.exception(0, True)

    with swallowing_raises(Exception, match="THIS EXCEPTION IS EXPECTED"):
        f.run(fail_fast=True)

    with Flow("flow") as f:
        with Stage("failure_stage"):
            x = m.exception(0, True)
            m.noop(x)

    with swallowing_raises(Exception, match="THIS EXCEPTION IS EXPECTED"):
        f.run(fail_fast=True)


def test_return_pipedag_config():
    with Flow("flow") as f:
        with Stage("failure_stage"):
            m.noop(PipedagConfig.default)

    with swallowing_raises(TypeError, match="PipedagConfig"):
        f.run(fail_fast=True)


def test_return_config_context():
    with Flow("flow") as f:
        with Stage("failure_stage"):
            m.noop(PipedagConfig.default.get())

    with swallowing_raises(TypeError, match="ConfigContext"):
        f.run(fail_fast=True)


def test_materialize_memo_literal():
    # A flow should be able to contain the same task with the same inputs
    # more than once and still run successfully.

    with Flow("flow") as f:
        with Stage("stage_0"):
            t_0 = m.noop(1)
            t_1 = m.noop(1)

            one = m.one()
            t_2 = m.noop(one)
            t_3 = m.noop(one)

        with Stage("stage_1"):
            t_4 = m.noop(1)
            t_5 = m.noop(t_4)
            t_6 = m.noop(t_5)
            t_7 = m.noop(t_0)

        with Stage("stage_2"):
            m.assert_equal(t_0, t_4)
            m.assert_equal(t_1, t_5)
            m.assert_equal(t_2, t_6)
            m.assert_equal(t_3, t_7)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_memo_table(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage_0"):
            t_0 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_1 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_2 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_3 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})

            _m.assert_table_equal(t_0, t_1)
            _m.assert_table_equal(t_1, t_2)
            _m.assert_table_equal(t_2, t_3)

        with Stage("stage_1"):
            t_4 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_5 = _m.noop(t_4)
            t_6 = _m.noop(t_5)
            t_7 = _m.noop(t_0)
            t_8 = _m.noop_sql(t_4)
            t_9 = _m.noop_sql(t_5)
            t_10 = _m.noop_sql(t_0)
            t_11 = _m.noop_lazy(t_4)
            t_12 = _m.noop_lazy(t_5)
            t_13 = _m.noop_lazy(t_0)
            if _m.pl is not None:
                t_14 = _m.noop_polars(t_4)
                t_15 = _m.noop_polars(t_5)
                t_16 = _m.noop_polars(t_0)
                t_17 = _m.noop_lazy_polars(t_4)
                t_18 = _m.noop_lazy_polars(t_5)
                t_19 = _m.noop_lazy_polars(t_0)

        with Stage("stage_2"):
            _m.assert_table_equal(t_0, t_4)
            _m.assert_table_equal(t_1, t_5)
            _m.assert_table_equal(t_2, t_6)
            _m.assert_table_equal(t_3, t_7)
            _m.assert_table_equal(t_1, t_8)
            _m.assert_table_equal(t_2, t_9)
            _m.assert_table_equal(t_3, t_10)
            _m.assert_table_equal(t_1, t_11)
            _m.assert_table_equal(t_2, t_12)
            _m.assert_table_equal(t_3, t_13)
            if _m.pl is not None:
                _m.assert_table_equal(t_1, t_14)
                _m.assert_table_equal(t_2, t_15)
                _m.assert_table_equal(t_3, t_16)
                _m.assert_table_equal(t_1, t_17)
                _m.assert_table_equal(t_2, t_18)
                _m.assert_table_equal(t_3, t_19)

    assert f.run().successful


def test_materialize_memo_with_failure():
    with Flow("flow") as f:
        with Stage("stage"):
            m.exception(1, False)
            m.exception(1, False)

            one = m.one()
            m.exception(1, True)
            t_3 = m.exception(one, True)
            m.exception(t_3, False)

    with ConfigContext.get().evolve(swallow_exceptions=True):
        assert not f.run(fail_fast=False).successful


def test_stage_ref_counter():
    def check_rc(stage, expected):
        @materialize(lazy=True, input_type=pd.DataFrame)
        def _m_check_rc(x):
            assert RunContext.get().get_stage_ref_count(stage) == expected
            return x

        return _m_check_rc

    # Super simple case with just two task inside one stage
    with Flow("flow") as f:
        with Stage("stage") as s:
            task_1 = m.one()
            task_2 = m.two()

            # One reference from the check_rc and commit
            check_rc(s, 2)([task_1, task_2])

    assert f.run().successful

    # Multiple tasks with interdependency inside one stage
    with Flow("flow") as f:
        with Stage("stage") as s:
            task_1 = m.one()
            task_2 = m.two()
            task_tuple = m.noop((task_1, task_2))
            # One reference from assert, noop, assert and commit
            task_tuple = check_rc(s, 4)(task_tuple)
            task_tuple = m.noop(task_tuple)
            # One reference from assert and commit
            check_rc(s, 2)(task_tuple)

    assert f.run().successful

    # Multiple tasks spread over multiple stages
    with Flow("flow") as f:
        with Stage("stage_1") as s1:
            task_1 = m.one()
            # One reference from assert, noop, assert, commit and downstream
            task_1 = check_rc(s1, 5)(task_1)
            task_1 = m.noop(task_1)
            # One reference from assert, commit and downstream
            check_rc(s1, 3)(task_1)

        with Stage("stage_2") as s2:
            task_2 = m.two()
            task_2 = m.noop([task_2])
            check_rc(s2, 3)(task_2)

        with Stage("stage_3") as s3:
            task_tuple = m.noop((task_1, task_2))
            # Check that s1 and s2 have been freed
            x = check_rc(s1, 0)(task_tuple)
            x = check_rc(s2, 0)(x)
            check_rc(s3, 2)(x)

    assert f.run().successful


def test_nout_and_getitem():
    with Flow("flow") as f:
        with Stage("stage_0"):
            x = m.noop((0, 1))
            zero, one = m.create_tuple(x[0], x[1])

            m.assert_equal(zero, 0)
            m.assert_equal(one, m.one())

        with Stage("stage_1"):
            d = m.noop({"a": [0, 1, 2, 3], "b": [3, 2, 1, 0]})

            a = d["a"]
            b = d["b"]

            for i in range(4):
                m.assert_equal(a[i], i)
                m.assert_equal(a[i], b[3 - i])

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_name_mangling_tables(imperative):
    @materialize
    def table_task_1():
        tbl = Table(pd.DataFrame({"x": [1]}), name="table_%%")
        return tbl.materialize() if imperative else tbl

    @materialize
    def table_task_2():
        tbl = Table(pd.DataFrame({"x": [2]}), name="table_%%")
        return tbl.materialize() if imperative else tbl

    with Flow() as f:
        with Stage("stage_1"):
            table_1 = table_task_1()
            table_2 = table_task_2()

    with StageLockContext():
        result = f.run()
        assert result.get(table_1, as_type=pd.DataFrame)["x"][0] == 1
        assert result.get(table_2, as_type=pd.DataFrame)["x"][0] == 2


@pytest.mark.parametrize("imperative", [False, True])
def test_name_mangling_lazy_tables(imperative):
    @materialize(lazy=True)
    def lazy_task_1():
        tbl = Table(select_as(1, "x"), name="table_%%")
        return tbl.materialize() if imperative else tbl

    @materialize(lazy=True)
    def lazy_task_2():
        tbl = Table(select_as(2, "x"), name="table_%%")
        return tbl.materialize() if imperative else tbl

    with Flow() as f:
        with Stage("stage_1"):
            lazy_1 = lazy_task_1()
            lazy_2 = lazy_task_2()

    with StageLockContext():
        result = f.run()
        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1
        assert result.get(lazy_2, as_type=pd.DataFrame)["x"][0] == 2


@pytest.mark.parametrize("imperative", [False, True])
def test_name_mangling_lazy_table_cache_fn(imperative):
    # Only the cache fn output of these two tasks is different
    @materialize(lazy=True, cache=lambda: 1, name="lazy_task")
    def lazy_task_1():
        tbl = Table(select_as(1, "x"))
        return tbl.materialize() if imperative else tbl

    @materialize(lazy=True, cache=lambda: 2, name="lazy_task")
    def lazy_task_2():
        tbl = Table(select_as(2, "x"))
        return tbl.materialize() if imperative else tbl

    with Flow() as f:
        with Stage("stage_1"):
            lazy_1 = lazy_task_1()
            lazy_2 = lazy_task_2()

    with StageLockContext():
        result = f.run()
        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1
        assert result.get(lazy_2, as_type=pd.DataFrame)["x"][0] == 2


def test_run_flow_with_empty_stage():
    with Flow("f") as f:
        with Stage("stage_0"):
            pass

    assert f.run().successful


@materialize(lazy=True, input_type=sa.Table)
def _check_nullable_task(tables):
    assert len(tables) == 18
    # This is dialect specific:
    # assert [c.nullable for c in tables[0].c] == [True, True, True]
    assert [c.nullable for c in tables[1].c] == [False, True, False]
    assert [c.nullable for c in tables[2].c] == [True, False, True]
    assert [c.nullable for c in tables[3].c] == [True, False, True]
    assert [c.nullable for c in tables[4].c] == [False, True, False]
    assert [c.nullable for c in tables[5].c] == [True, True, True]
    assert [c.nullable for c in tables[6].c] == [True, False, True]
    assert [c.nullable for c in tables[7].c] == [False, True, False]
    assert [c.nullable for c in tables[8].c] == [False, False, False]
    # assert [c.nullable for c in tables[9].c] == [True, True, True]
    assert [c.nullable for c in tables[10].c][1:] == [True, False]
    assert [c.nullable for c in tables[11].c][1:] == [False, True]
    assert [c.nullable for c in tables[12].c][1:] == [False, True]
    assert [c.nullable for c in tables[13].c][1:] == [True, False]
    assert [c.nullable for c in tables[14].c][1:] == [True, True]
    assert [c.nullable for c in tables[15].c][1:] == [False, True]
    assert [c.nullable for c in tables[16].c][1:] == [True, False]
    assert [c.nullable for c in tables[17].c][1:] == [False, False]
    return tables


def _test_nullable_lazy_get_flow():
    cols = sa.literal(1).label("x"), sa.literal(2).label("y"), sa.literal(3).label("z")

    @materialize(lazy=True)
    def lazy_task_1():
        q = sa.select(*cols)
        tables = [
            Table(q),
            Table(q, nullable=["y"]),
            Table(q, nullable=["x", "z"]),
            Table(q, non_nullable=["y"]),
            Table(q, non_nullable=["x", "z"]),
            Table(q, nullable=["x", "y", "z"], non_nullable=[]),
            Table(q, nullable=["x", "z"], non_nullable=["y"]),
            Table(q, nullable=["y"], non_nullable=["x", "z"]),
            Table(q, nullable=[], non_nullable=["x", "y", "z"]),
            Table(q, primary_key=["x"]),
            Table(q, primary_key=["x"], nullable=["y"]),
            Table(q, primary_key=["x"], nullable=["x", "z"]),
            Table(q, primary_key=["x"], non_nullable=["y"]),
            Table(q, primary_key=["x"], non_nullable=["x", "z"]),
            Table(q, primary_key=["x"], nullable=["x", "y", "z"], non_nullable=[]),
            Table(q, primary_key=["x"], nullable=["x", "z"], non_nullable=["y"]),
            Table(q, primary_key=["x"], nullable=["y"], non_nullable=["x", "z"]),
            Table(q, primary_key=["x"], nullable=[], non_nullable=["x", "y", "z"]),
        ]
        return tables

    with Flow() as f:
        with Stage("stage_1"):
            lazy_tables = lazy_task_1()
            _check_nullable_task(lazy_tables)
    return f, lazy_tables


def _test_nullable_get_flow():
    @materialize
    def task_1():
        df = pd.DataFrame({"x": [1], "y": [2], "z": [3]})
        tables = [
            Table(df),
            Table(df, nullable=["y"]),
            Table(df, nullable=["x", "z"]),
            Table(df, non_nullable=["y"]),
            Table(df, non_nullable=["x", "z"]),
            Table(df, nullable=["x", "y", "z"], non_nullable=[]),
            Table(df, nullable=["x", "z"], non_nullable=["y"]),
            Table(df, nullable=["y"], non_nullable=["x", "z"]),
            Table(df, nullable=[], non_nullable=["x", "y", "z"]),
            Table(df, primary_key=["x"]),
            Table(df, primary_key=["x"], nullable=["y"]),
            Table(df, primary_key=["x"], nullable=["x", "z"]),
            Table(df, primary_key=["x"], non_nullable=["y"]),
            Table(df, primary_key=["x"], non_nullable=["x", "z"]),
            Table(df, primary_key=["x"], nullable=["x", "y", "z"], non_nullable=[]),
            Table(df, primary_key=["x"], nullable=["x", "z"], non_nullable=["y"]),
            Table(df, primary_key=["x"], nullable=["y"], non_nullable=["x", "z"]),
            Table(df, primary_key=["x"], nullable=[], non_nullable=["x", "y", "z"]),
        ]
        return tables

    with Flow() as f:
        with Stage("stage_1"):
            lazy_tables = task_1()
            _check_nullable_task(lazy_tables)
    return f, lazy_tables


@skip_instances("snowflake", "parquet_backend", "parquet_s3_backend", "parquet_s3_backend_db2")
@pytest.mark.parametrize("_get_flow", [_test_nullable_get_flow, _test_nullable_lazy_get_flow])
def test_nullable(_get_flow):
    f, tables = _get_flow()

    with StageLockContext():
        result = f.run()
        assert result.successful
        for i in range(18):  # unfortunately, 'for tbl in tables' does not work
            tbl = tables[i]
            assert len(result.get(tbl, as_type=pd.DataFrame)) == 1
            assert result.get(tbl, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(tbl, as_type=pd.DataFrame)["y"][0] == 2
            assert result.get(tbl, as_type=pd.DataFrame)["z"][0] == 3
        # check that there are 9 tables with one primary key column each by checking
        # the sum of all primary key columns
        if ConfigContext.get().store.table_store.engine.dialect.name != "duckdb":
            assert (
                sum(col.primary_key for i in range(18) for col in result.get(tables[i], as_type=sa.Table).original.c)
                == 9
            )


# duckdb does not persist nullable flags over connections; snowflake is slow
@skip_instances("duckdb", "snowflake", "parquet_backend", "parquet_s3_backend", "parquet_s3_backend_db2")
@pytest.mark.parametrize("_get_flow", [_test_nullable_get_flow, _test_nullable_lazy_get_flow])
def test_nullable_output(_get_flow):
    f, tables = _get_flow()

    def get_nullable(idx):
        return [c.nullable for c in result.get(tables[idx], as_type=sa.Table).c]

    with StageLockContext():
        result = f.run()
        assert result.successful
        # This is dialect specific:
        # assert get_nullable(0) == [True, True, True]
        assert get_nullable(1) == [False, True, False]
        assert get_nullable(2) == [True, False, True]
        assert get_nullable(3) == [True, False, True]
        assert get_nullable(4) == [False, True, False]
        assert get_nullable(5) == [True, True, True]
        assert get_nullable(6) == [True, False, True]
        assert get_nullable(7) == [False, True, False]
        assert get_nullable(8) == [False, False, False]


@pytest.mark.parametrize(
    "nullable, non_nullable, error",
    [
        (["a"], None, ValueError),
        (None, ["a"], ValueError),
        (["a", "x", "y", "z"], None, ValueError),
        (None, ["a", "x", "y", "z"], ValueError),
        (["x", "y", "z"], ["a"], ValueError),
        (["a"], ["x", "y", "z"], ValueError),
        (["x"], ["x"], ValueError),
        (["x"], ["y"], ValueError),
        (["x", "y"], ["y"], ValueError),
        ("x", None, TypeError),
        (None, "x", TypeError),
        (1, None, TypeError),
        (None, 2, TypeError),
    ],
)
def test_nullable_raises(nullable, non_nullable, error):
    cols = sa.literal(1).label("x"), sa.literal(2).label("y"), sa.literal(3).label("z")

    @materialize(lazy=True)
    def lazy_task_1():
        q = sa.select(*cols)
        df = pd.DataFrame({"x": [1], "y": [2], "z": [3]})
        tables = [
            Table(q, nullable=nullable, non_nullable=non_nullable),
            Table(df, nullable=nullable, non_nullable=non_nullable),
            Table(q, primary_key=["x"], nullable=nullable, non_nullable=non_nullable),
            Table(df, primary_key=["x"], nullable=nullable, non_nullable=non_nullable),
        ]
        return tables

    with Flow() as f:
        with Stage("stage_1"):
            lazy_task_1()

    with swallowing_raises(error):
        f.run()


@materialize(lazy=True)
def _lazy_task_1():
    return Table(select_as(1, "x"), name="t1", primary_key="x")


@materialize(lazy=True, nout=2)
def _lazy_task_2():
    return Table(select_as(1, "x"), name="t21", indexes=[["x"]]), Table(
        select_as(2, "x"), name="t22", primary_key=["x"]
    )


@materialize(lazy=True, input_type=sa.Table)
def _lazy_join(src1: sa.sql.expression.Alias, src2: sa.sql.expression.Alias):
    query = sa.select(src1.c.x, src2.c.x.label("x2")).select_from(src1.outerjoin(src2, src1.c.x == src2.c.x))
    return Table(query, "t3_%%", indexes=[["x2"], ["x", "x2"]])


@materialize(version="1.0")
def _sql_task_1():
    return Table(select_as(1, "x"), name="t1", primary_key="x")


@materialize(version="1.0", nout=2)
def _sql_task_2():
    return (
        Table(select_as(1, "x"), name="t21", indexes=[["x"]]),
        Table(select_as(2, "x"), name="t22", primary_key=["x"]),
    )


@materialize(version="1.0", input_type=sa.Table)
def _sql_join(src1: sa.sql.expression.Alias, src2: sa.sql.expression.Alias):
    query = sa.select(src1.c.x, src2.c.x.label("x2")).select_from(src1.outerjoin(src2, src1.c.x == src2.c.x))
    return Table(query, "t3_%%", indexes=[["x2"], ["x", "x2"]])


@materialize(version="1.0")
def _eager_task_1():
    df = pd.DataFrame(dict(x=[1]))
    return Table(df, name="t1", primary_key="x")


@materialize(version="1.0", nout=2)
def _eager_task_2():
    df1 = pd.DataFrame(dict(x=[1]))
    df2 = pd.DataFrame(dict(x=[2]))
    return (
        Table(df1, name="t21", indexes=[["x"]]),
        Table(df2, name="t22", primary_key=["x"]),
    )


@materialize(version="1.0", input_type=pd.DataFrame)
def _eager_join(src1: pd.DataFrame, src2: pd.DataFrame):
    src2["x2"] = src2["x"]
    join = src1.merge(src2, on="x", how="left")
    return Table(join, "t3_%%", indexes=[["x2"], ["x", "x2"]])


@pytest.mark.skip
@pytest.mark.parametrize(
    "task_1,task_2,noop,join",
    [
        # communicate with same kind
        (_lazy_task_1, _lazy_task_2, m.noop_lazy, _lazy_join),
        (_sql_task_1, _sql_task_2, m.noop_sql, _sql_join),
        (_eager_task_1, _eager_task_2, m.noop, _eager_join),
        # communicate to sql
        (_lazy_task_1, _lazy_task_2, m.noop_sql, _sql_join),
        (_eager_task_1, _eager_task_2, m.noop_sql, _sql_join),
        # communicate to lazy
        (_sql_task_1, _sql_task_2, m.noop_lazy, _lazy_join),
        (_eager_task_1, _eager_task_2, m.noop_lazy, _lazy_join),
        # communicate to eager
        (_lazy_task_1, _lazy_task_2, m.noop, _eager_join),
        (_sql_task_1, _sql_task_2, m.noop, _eager_join),
    ],
)
def test_task_and_stage_communication(task_1, task_2, noop, join):
    logger = structlog.get_logger("test_task_and_stage_communication")
    with Flow() as f:
        with Stage("stage_1"):
            t1 = task_1()
            t21, t22 = task_2()
            same_stage = noop(t22)
            same_stage2 = join(t1, t21)
        with Stage("stage_2"):
            next_stage = noop(t22)
            next_stage2 = join(t1, t21)
            next_same_stage = join(t1, next_stage)

    for i in range(3):
        with StageLockContext():
            logger.info("## Running flow", iteration=i)
            result = f.run()
            logger.info("#    checking flow", iteration=i)
            assert result.successful
            assert result.get(t1, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(t21, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(t22, as_type=pd.DataFrame)["x"][0] == 2
            assert result.get(same_stage, as_type=pd.DataFrame)["x"][0] == 2
            assert result.get(same_stage2, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(same_stage2, as_type=pd.DataFrame)["x2"][0] == 1
            assert result.get(next_stage, as_type=pd.DataFrame)["x"][0] == 2
            assert result.get(next_stage2, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(next_stage2, as_type=pd.DataFrame)["x2"][0] == 1
            assert result.get(next_same_stage, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(next_same_stage, as_type=pd.DataFrame)["x2"].isna().all()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import dataframely as dy
import pandas as pd
import polars as pl
import sqlalchemy as sa

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import AUTO_VERSION, Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext


@materialize(lazy=True)
def lazy_task_1():
    return Table(
        sa.select(
            sa.literal(1).label("x"),
            sa.literal(2).label("y"),
        ),
        name="lazy_1",
    )


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_2(input1: sa.Alias, input2: sa.Alias):
    query = sa.select(
        (input1.c.x * 5).label("x5"),
        input2.c.a,
    ).select_from(input1.outerjoin(input2, input2.c.x == input1.c.x))

    return Table(query, name="task_2_out", primary_key=["a"])


def ref(tbl: sa.Alias):
    # For case sensitive (mixed capital/lowercase) names, quoting is necessary.
    # But it is recommended to avoid using this with Snowflake since the default is uppercase.
    return f"{tbl.original.schema}.{tbl.original.name}"


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_3(input1: sa.Alias):
    # With ref() it is recommended to use table alias in the query since the actual table name might change
    return sa.text(f"SELECT input1.* FROM {ref(input1)} as input1")


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_4(input1: sa.Alias):
    return sa.text(f"SELECT * FROM {ref(input1)}")


@materialize(nout=2, version="1.0.0")
def eager_inputs():
    dfA = pd.DataFrame(
        {
            "a": [0, 1, 2, 4],
            "b": [9, 8, 7, 6],
        }
    )
    dfB = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(dfA, "dfA"), Table(dfB, "dfB_%%")


@materialize(version="1.0.0", input_type=pd.DataFrame)
def eager_task(tbl1: pd.DataFrame, tbl2: pd.DataFrame):
    return tbl1.merge(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_polars(tbl1: pl.LazyFrame, tbl2: pl.LazyFrame):
    return tbl1.join(tbl2, on="x")


class Tbl1Schema(dy.Schema):
    x = dy.Int32()
    y = dy.Int32()


class OutputSchema(Tbl1Schema):
    a = dy.Int64()


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_dataframely(tbl1: dy.LazyFrame[Tbl1Schema], tbl2: pl.LazyFrame) -> OutputSchema:
    # Pipedag automatically calls Tbl1Schema.cast() and OutputSchema.validate().
    return tbl1.join(tbl2, on="x")


def main():
    with Flow() as f:
        with Stage("stage_1"):
            lazy_1 = lazy_task_1()
            a, b = eager_inputs()

        with Stage("stage_2"):
            lazy_2 = lazy_task_2(lazy_1, b)
            lazy_3 = lazy_task_3(lazy_2)
            # Do the same operation with various syntax alternatives.
            # This is also a demonstration how pipedag interacts with \
            # dataframely.
            out = eager_task(lazy_1, b)
            out2 = eager_task_polars(lazy_1, b)
            out5 = eager_task_dataframely(lazy_1, b)

        with Stage("stage_3"):
            lazy_4 = lazy_task_4(lazy_2)
        _ = lazy_3, lazy_4, out, out2, out5  # unused terminal output tables

    # Run flow
    result = f.run()
    assert result.successful

    # Run in a different way for testing
    with StageLockContext():
        result = f.run()
        assert result.successful
        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish

    # Run this pipeline with (might take a bit longer on first run in pixi environment):
    # ```shell
    # export SNOWFLAKE_ACCOUNT=<use your snowflake instance>;
    # export SNOWFLAKE_PASSWORD=<use secret token>;
    # export SNOWFLAKE_USER=<username>;
    # export SNOWFLAKE_DB_SUFFIX=_adhoc;  # chose unique suffix to avoid collisions
    # pixi run python run_pipeline.py
    # ```

    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import types
from typing import Generic, TypeVar

try:
    import dataframely as dy
    from dataframely._base_schema import SchemaMeta
    from dataframely._polars import FrameType
    from dataframely.random import Generator
    from dataframely.testing import validation_mask
except ImportError:

    class Generator:
        pass

    T = TypeVar("T")

    class DyDataFrame(Generic[T]):
        pass

    class DyDummyClass:
        def __init__(self, *args, **kwargs):
            pass

    FrameType = None
    SchemaMeta = None
    validation_mask = None
    dy = types.ModuleType("dataframely")
    dy.DataFrame = DyDataFrame
    dy.LazyFrame = DyDataFrame
    dy.FailureInfo = None
    dy.Column = None
    dy.Collection = object
    dy.Schema = DyDummyClass
    dy.filter = lambda: lambda fn: fn  # noqa
    dy.rule = lambda: lambda fn: fn  # noqa
    for _type in [
        "Int8",
        "Int16",
        "Int32",
        "Int64",
        "UInt8",
        "UInt16",
        "UInt32",
        "UInt64",
        "Float32",
        "Float64",
        "Bool",
        "String",
        "Decimal",
        "Enum",
        "Struct",
        "List",
        "Date",
        "Datetime",
        "Time",
        "Duration",
        "Float",
        "Integer",
    ]:
        setattr(dy, _type, DyDummyClass)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

try:
    import kazoo
    from kazoo.client import KazooClient, KazooState
    from kazoo.recipe.lock import Lock as KazooLock
except ImportError:
    kazoo = None
    KazooClient = None
    KazooState = None
    KazooLock = None
